import copy
import dataclasses
import json

import numpy as np

from constants.defaults import DEFAULT_NAN_VALUE, PRINT_QUERIES
from constants.methods import factory
from database.Database import Database
from database.Operators import Operators
from entities.Feature import Feature
from entities.Record import Record
from entities.Resource import Resource
from enums.DataTypes import DataTypes
from enums.Profile import Profile
from enums.TableNames import TableNames
from utils.setup_logger import log


@dataclasses.dataclass()
class FeatureProfileComputation:
    database: Database

    def __post_init__(self):
        # store, for each dataset, the total number of patients to compute the missing percentage
        operations = [
            Operators.match(field=None,
                            value={Resource.ENTITY_TYPE_: {"$ne": f"{Profile.CLINICAL}{TableNames.RECORD}"}},
                            is_regex=False),
            Operators.group_by(group_key={Record.DATASET_: Record.DATASET__, Record.SUBJECT_: Record.SUBJECT__},
                               groups=[{"name": "nb_patients", "operator": "$sum", "field": 1}]),
            Operators.group_by(group_key={Record.DATASET_: Record.DATASET___},
                               groups=[{"name": "distinct_nb_patients", "operator": "$sum", "field": 1}]),
            Operators.write_to_table(table_name=TableNames.COUNTS_PATIENTS)
        ]
        if PRINT_QUERIES:
            log.info(operations)
        _ = self.database.db[TableNames.RECORD].aggregate(operations)

        # store, for each clinical dataset, the total number of samples to compute the missing percentage for clinical data
        operations = [
            Operators.match(field=None,
                            value={Resource.ENTITY_TYPE_: {"$eq": f"{Profile.CLINICAL}{TableNames.RECORD}"}},
                            is_regex=False),
            Operators.group_by(group_key={Record.DATASET_: Record.DATASET__, Record.BASE_ID_: Record.BASE_ID__},
                               groups=[{"name": "nb_samples", "operator": "$sum", "field": 1}]),
            Operators.group_by(group_key={Record.DATASET_: Record.DATASET___},
                               groups=[{"name": "distinct_nb_samples", "operator": "$sum", "field": 1}]),
            Operators.write_to_table(table_name=TableNames.COUNTS_SAMPLES)
        ]
        if PRINT_QUERIES:
            log.info(operations)
        _ = self.database.db[TableNames.RECORD].aggregate(operations)

        # compute the list of Feature identifiers for each Profile data type (numeric, category, date)
        # the filter {"datasets": self.dataset_gid} means that the array "datasets" contains the element self.dataset_gid
        cursor = self.database.find_operation(table_name=TableNames.FEATURE, filter_dict={},
                                              projection={Resource.IDENTIFIER_: 1, Feature.DT_: 1, "_id": 0})
        map_feature_datatype = {}
        for element in cursor:
            if element[Feature.DT_] not in map_feature_datatype:
                map_feature_datatype[element[Feature.DT_]] = []
            map_feature_datatype[element[Feature.DT_]].append(element[Resource.IDENTIFIER_])
        # log.info(map_feature_datatype)

        self.numeric_features = []
        self.numeric_features.extend(
            map_feature_datatype[DataTypes.INTEGER] if DataTypes.INTEGER in map_feature_datatype else [])
        self.numeric_features.extend(
            map_feature_datatype[DataTypes.FLOAT] if DataTypes.FLOAT in map_feature_datatype else [])
        self.categorical_features = []
        self.categorical_features.extend(
            map_feature_datatype[DataTypes.STRING] if DataTypes.STRING in map_feature_datatype else [])
        self.categorical_features.extend(
            map_feature_datatype[DataTypes.BOOLEAN] if DataTypes.BOOLEAN in map_feature_datatype else [])
        self.categorical_features.extend(
            map_feature_datatype[DataTypes.CATEGORY] if DataTypes.CATEGORY in map_feature_datatype else [])
        self.date_features = []
        self.date_features.extend(
            map_feature_datatype[DataTypes.DATE] if DataTypes.DATE in map_feature_datatype else [])
        self.date_features.extend(
            map_feature_datatype[DataTypes.DATETIME] if DataTypes.DATETIME in map_feature_datatype else [])
        self.all_features = self.numeric_features + self.categorical_features + self.date_features

    def compute_features_profiles(self) -> None:
        # clear FeatureProfile table and create a unique index on <dataset, instantiates>
        # because merge requires a unique index on the merge keys
        self.database.drop_table(TableNames.FEATURE_PROFILE)
        self.database.create_unique_index(TableNames.FEATURE_PROFILE,
                                          columns={Record.DATASET_: 1, Record.INSTANTIATES_: 1})

        # ALL FEATURES
        operators = self.data_type_validity_query(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # NUMERIC FEATURES
        match_numeric_values = [Operators.match(field=None, value=Operators.or_operator([
            {Record.VALUE_: {"$type": "int"}},
            {Record.VALUE_: {"$type": "double"}},
            {Record.VALUE_: {"$type": "long"}},
            {Record.VALUE_: {"$type": "decimal"}}]), is_regex=False),
                                Operators.match(field=None, value={Record.VALUE_: {"$ne": np.nan}}, is_regex=False)]

        # 1. compute min, max, mean, median, std values for numerical features
        operators = copy.deepcopy(match_numeric_values)
        operators.extend(
            self.min_max_mean_median_std_query(features_ids=self.numeric_features, compute_min=True, compute_max=True,
                                               compute_mean=True, compute_median=True, compute_std=True))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 2. compute the Median Absolute Deviation
        operators = copy.deepcopy(match_numeric_values)
        operators.extend(self.abs_med_dev_query(features_ids=self.numeric_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 3. compute skewness and kurtosis for numerical features
        operators = copy.deepcopy(match_numeric_values)
        operators.extend(self.skewness_and_kurtosis_query(features_ids=self.numeric_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 4. compute IQR for numerical features
        operators = copy.deepcopy(match_numeric_values)
        operators.extend(self.iqr_query(features_ids=self.numeric_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 5. compute Pearson correlation coefficients
        operators = copy.deepcopy(match_numeric_values)
        operators.extend(self.pearson_correlation_query(features_ids=self.numeric_features, database=self.database))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.PAIRS_FEATURES].aggregate(operators)

        # DATE FEATURES
        match_date_values = [Operators.match(field=None, value=Operators.or_operator([
            {Record.VALUE_: {"$type": "date"}},
            {Record.VALUE_: {"$type": "timestamp"}}]), is_regex=False),
                             Operators.match(field=None, value={Record.VALUE_: {"$ne": np.nan}}, is_regex=False)]

        # 1. compute min, max, mean, median, std values for date features
        operators = copy.deepcopy(match_date_values)
        operators.extend(
            self.min_max_mean_median_std_query(features_ids=self.date_features, compute_min=True, compute_max=True,
                                               compute_mean=False, compute_median=True, compute_std=False))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        # self.database.db[TableNames.RECORD].aggregate(operators)

        # 2. compute IQR for numerical features
        operators = copy.deepcopy(match_date_values)
        operators.extend(self.iqr_query(features_ids=self.date_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        # self.database.db[TableNames.RECORD].aggregate(operators)

        # CATEGORICAL FEATURES
        match_categorical_values = [Operators.match(field=None, value=Operators.or_operator([
            {Record.VALUE_: {"$type": "object"}}]), is_regex=False),
                                    Operators.match(field=None, value={Record.VALUE_: {"$ne": np.nan}}, is_regex=False)]

        # 1. compute imbalance for categorical features
        operators = copy.deepcopy(match_categorical_values)
        operators.extend(self.imbalance_query(features_ids=self.categorical_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 2. compute constancy for categorical features
        operators = copy.deepcopy(match_categorical_values)
        operators.extend(self.constancy_query(features_ids=self.categorical_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # 3. compute mode for categorical features
        operators = copy.deepcopy(match_categorical_values)
        operators.extend(self.mode_query(features_ids=self.categorical_features))
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # SHARED PROFILE FEATURES
        # uniqueness
        operators = self.uniqueness_query(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # entropy
        operators = self.entropy_query(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # density
        operators = self.density_query(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        # values and counts
        operators = self.values_and_counts_query(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=True))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        operators = self.missing_percentage_query_without_samples(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

        operators = self.missing_percentage_query_with_samples(features_ids=self.all_features)
        operators.extend(self.finalize_query(include_value=False))
        if PRINT_QUERIES:
            log.info(operators)
        self.database.db[TableNames.RECORD].aggregate(operators)

    def min_max_mean_median_std_query(self, features_ids: list, compute_min: bool, compute_max: bool,
                                      compute_mean: bool, compute_median: bool, compute_std: bool) -> list:
        groups = []
        if compute_min:
            groups.append({"name": "min_value", "operator": "$min", "field": Record.VALUE__})
        if compute_max:
            groups.append({"name": "max_value", "operator": "$max", "field": Record.VALUE__})
        if compute_mean:
            groups.append({"name": "mean_value", "operator": "$avg", "field": Record.VALUE__})
        if compute_median:
            groups.append({"name": "median_value", "operator": "$median",
                           "field": {"input": Record.VALUE__, "method": "approximate"}})
        if compute_std:
            groups.append({"name": "std_value", "operator": "$stdDevPop", "field": Record.VALUE__})
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__},
                groups=groups),
        ]

    def abs_med_dev_query(self, features_ids: list) -> list:
        # i.e., EMA=median(|Xi-Y|) where Xi is a value, Y is the median, and L is the absolute value (no minus sign)
        # [{'$group': {'_id': {'_id': null}, 'originalValues': {'$push': '$value'}, 'mymedian': {'$median': {"input": "$value", "method": "approximate"}}}}, {'$unwind': '$originalValues'}, {'$project': {"absVal": {'$abs': {'$subtract': ['$originalValues', '$mymedian']}}}}, {'$group': {'_id': {'_id': null}, 'ema': {'$median': {"input": "$absVal", "method": "approximate"}}}}]
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "originalValues", "operator": "$push", "field": Record.VALUE__},
                    {"name": "mymedian", "operator": "$median",
                     "field": {"input": Record.VALUE__, "method": "approximate"}},
                ]),
            # after groupby, the access to individual elements is lost, so we need to use $push or $addToSet to keep track of original values
            Operators.unwind("originalValues"),
            Operators.project(field=None,
                              projected_value={"absVal": {"$abs": {"$subtract": ["$originalValues", "$mymedian"]}}}),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET___, Record.INSTANTIATES_: Record.INSTANTIATES___}, groups=[
                    {"name": "ema", "operator": "$median", "field": {"input": "$absVal", "method": "approximate"}}
                ])
        ]

    def skewness_and_kurtosis_query(self, features_ids: list) -> list:
        # kurtosis query
        # [ { "$group": { "_id": { "_id": null }, "originalValues": { "$push": "$value" }, "mymean": { "$avg": "$value" }, "mystdDev": { "$stdDevSamp": "$value" } } }, { "$unwind": "$originalValues" }, {"$project":{"thePowed":{"$pow":[{"$divide":[{"$subtract":["$originalValues","$mymean"]},"$mystdDev"]},4]}}},{"$group":{"_id":{"_id":null},"summed":{"$sum":"$thePowed"}, "summedValues": { "$push": "$thePowed" }}},{"$project":{"summed": 1, "summedValues": 1, "mysize": {"$size": "$summedValues"}}}, {"$project":{"kurtosis":{"$subtract":[{"$multiply":[{"$divide":[{"$multiply":["$mysize",{"$sum":["$mysize",1]}]},{"$multiply":[{"$sum":["$mysize",-1]},{"$sum":["$mysize",-2]},{"$sum":["$mysize",-3]}]}]},"$summed"]},{"$divide":[{"$multiply":[3,{"$pow":[{"$sum":["$mysize",-1]},2]}]},{"$multiply":[{"$sum":["$mysize",-2]},{"$sum":["$mysize",-3]}]}]}]}}}]
        # skewness query
        # [ { "$group": { "_id": { "_id": null }, "originalValues": { "$push": "$value" }, "mymean": { "$avg": "$value" }, "mystdDev": { "$stdDevSamp": "$value" } } }, { "$unwind": "$originalValues" }, {"$project":{"thePowed":{"$pow":[{"$divide":[{"$subtract":["$originalValues","$mymean"]},"$mystdDev"]},3]}}},{"$group":{"_id":{"_id":null},"summed":{"$sum":"$thePowed"}, "summedValues": { "$push": "$thePowed" }}},{"$project":{"summed": 1, "summedValues": 1, "mysize": {"$size": "$summedValues"}}}, {"$project":{"skewness":{"$multiply": [{"$divide": ["$mysize", {"$multiply": [{"$sum": ["$mysize", -1]}, {"$sum": ["$mysize", -2]}]}]}, "$summed"]}}}]
        # and we merge them to avoid repeating computation of mean, std, dev and the array with push
        # [ { "$group": { "_id": { "_id": null }, "originalValues": { "$push": "$value" }, "mymean": { "$avg": "$value" }, "mystdDev": { "$stdDevSamp": "$value" } } }, { "$unwind": "$originalValues" }, {"$project":{"theQuads":{"$pow":[{"$divide":[{"$subtract":["$originalValues","$mymean"]},"$mystdDev"]},4]}, "theSquares":{"$pow":[{"$divide":[{"$subtract":["$originalValues","$mymean"]},"$mystdDev"]},3]}}},{"$group":{"_id":{"_id":null},"sumQuads":{"$sum":"$theQuads"}, "sumSquares":{"$sum":"$theSquares"}, "summedValues": { "$push": "$theQuads" }}},{"$project":{"sumQuads": 1, "sumSquares": 1, "mysize": {"$size": "$summedValues"}}}, {"$project":{"kurtosis":{"$subtract":[{"$multiply":[{"$divide":[{"$multiply":["$mysize",{"$sum":["$mysize",1]}]},{"$multiply":[{"$sum":["$mysize",-1]},{"$sum":["$mysize",-2]},{"$sum":["$mysize",-3]}]}]},"$sumQuads"]},{"$divide":[{"$multiply":[3,{"$pow":[{"$sum":["$mysize",-1]},2]}]},{"$multiply":[{"$sum":["$mysize",-2]},{"$sum":["$mysize",-3]}]}]}]}, "skewness":{"$multiply": [{"$divide": ["$mysize", {"$multiply": [{"$sum": ["$mysize", -1]}, {"$sum": ["$mysize", -2]}]}]}, "$sumSquares"]}}}]
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "originalValues", "operator": "$push", "field": Record.VALUE__},
                    {"name": "mymean", "operator": "$avg", "field": Record.VALUE__},
                    {"name": "mystdDev", "operator": "$stdDevSamp", "field": Record.VALUE__}
                ]),
            # after groupby, the access to individual elements is lost, so we need to use $push or $addToSet to keep track of original values
            Operators.unwind("originalValues"),
            # to be able to apply $subtract to each element of the array -- this adds the $ in front of the variable name
            Operators.project(field=None, projected_value={
                "theQuads": {
                    "$cond": {
                        "if": {"$eq": ["$mystdDev", 0]},
                        "then": [0],
                        "else": {
                            "$pow": [{"$divide": [{"$subtract": ["$originalValues", "$mymean"]}, "$mystdDev"]}, 4]
                        }
                    }
                },
                "theSquares": {
                    "$cond": {
                        "if": {"$eq": ["$mystdDev", 0]},
                        "then": [0],
                        "else": {
                            "$pow": [{"$divide": [{"$subtract": ["$originalValues", "$mymean"]}, "$mystdDev"]}, 3]
                        }
                    }
                }}),
            Operators.group_by(group_key="$_id", groups=[
                {"name": "sumQuads", "operator": "$sum", "field": "$theQuads"},
                {"name": "sumSquares", "operator": "$sum", "field": "$theSquares"},
                {"name": "summedValues", "operator": "$push", "field": "$theQuads"}
                # to later compute the number of values, we keep track of all the elements that have been summed
            ]),
            Operators.project(field=None,
                              projected_value={"sumQuads": 1, "sumSquares": 1, "n": {"$size": "$summedValues"}}),
            Operators.project(field=None, projected_value={
                "kurtosis": {
                    "$cond": {
                        "if": {"$or": [
                            {"$eq": ["$n", 0]},
                            {"$eq": ["sumQuads", [0]]},
                            {"$eq": [{"$sum": ["$n", -1]}, 0]},
                            {"$eq": [{"$sum": ["$n", -2]}, 0]},
                            {"$eq": [{"$sum": ["$n", -3]}, 0]}
                        ]},
                        "then": 0,
                        "else": {
                            "$subtract": [
                                {"$multiply": [
                                    {"$divide": [
                                        {"$multiply": ["$n", {"$sum": ["$n", 1]}]},
                                        {"$multiply": [{"$sum": ["$n", -1]}, {"$sum": ["$n", -2]},
                                                       {"$sum": ["$n", -3]}]}
                                    ]},
                                    "$sumQuads"
                                ]}
                                , {"$divide": [
                                    {"$multiply": [3, {"$pow": [{"$sum": ["$n", -1]}, 2]}]},
                                    {"$multiply": [{"$sum": ["$n", -2]}, {"$sum": ["$n", -3]}]}
                                ]}
                            ]
                        }
                    }
                },
                "skewness": {
                    "$cond": {
                        "if": {"$or": [
                            {"$eq": ["$n", 0]},
                            {"$eq": ["sumSquares", [0]]},
                            {"$eq": [{"$sum": ["$n", -1]}, 0]},
                            {"$eq": [{"$sum": ["$n", -2]}, 0]}
                        ]},
                        "then": 0,
                        "else": {
                            "$multiply": [
                                {"$divide": [
                                    "$n", {"$multiply": [{"$sum": ["$n", -1]}, {"$sum": ["$n", -2]}]}
                                ]},
                                "$sumSquares"
                            ]
                        }
                    }
                }
            })
        ]

    def iqr_query(self, features_ids: list) -> list:
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "thevalues", "operator": "$push", "field": Record.VALUE__},
                    {"name": "themedian", "operator": "$median",
                     "field": {"input": Record.VALUE__, "method": "approximate"}}
                ]),
            Operators.set_variables(
                variables=[{"name": "thevalues", "operation": {"$sortArray": {"input": "$thevalues", "sortBy": 1}}}]),
            Operators.project(field=None, projected_value={"thevalues": {
                "$filter": {"input": "$thevalues", "as": "item", "cond": {"$ne": ["$$item", "$themedian"]}}}}),
            Operators.add_fields(key="thevalues", value={
                "$cond": {
                    "if": {"$eq": [{"$size": "$thevalues"}, 0]},
                    "then": [[], []],
                    "else": {
                        "$cond": {
                            "if": {"$eq": [{"mod": [{"$size": "$thevalues"}, 2]}, 0]},
                            "then": {
                                "$map": {
                                    "input": {"$range": [0, {"$size": "$thevalues"},
                                                         {"$divide": [{"$size": "$thevalues"}, 2]}]},
                                    "as": "index",
                                    "in": {"$slice": ["$thevalues", "$$index", 2]}
                                }
                            },
                            "else": {
                                "$map": {
                                    "input": {"$range": [0, {"$size": "$thevalues"},
                                                         {"$floor": {"$divide": [{"$size": "$thevalues"}, 2]}}]},
                                    "as": "index",
                                    "in": {"$slice": ["$thevalues", "$$index",
                                                      {"$floor": {"$divide": [{"$size": "$thevalues"}, 2]}}]}
                                }
                            }
                        }
                    }
                }
            }),
            Operators.set_variables(variables=[
                {"name": "q1",
                 "operation": {"$median": {"input": {"$arrayElemAt": ["$thevalues", 0]}, "method": "approximate"}}},
                {"name": "q3",
                 "operation": {"$median": {"input": {"$arrayElemAt": ["$thevalues", 1]}, "method": "approximate"}}},
            ]),
            Operators.set_variables(variables=[
                {"name": "iqr", "operation": {"$subtract": ["$q3", "$q1"]}}  # compute the IQR as Q3-Q1
            ]),
            Operators.unset_variables(["thevalues", "q1", "q3"])
        ]

    def pearson_correlation_query(self, features_ids: list, database: Database) -> list:
        # drop existing temporary tables
        database.drop_table(TableNames.TOP10_FEATURES)
        database.drop_table(TableNames.PAIRS_FEATURES)
        # first part: create pairs of features for which pearson coefficient will be computed
        # we only take the top-10 of numeric features, i.e.,
        # the 10 numeric features with most values (if tied, take the ones with the lowest ID)
        top_10_features = [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1}
                ]),
            Operators.sort(field=f"_id.{Record.DATASET_}", sort_order=-1),
            Operators.sort(field="frequency", sort_order=-1),
            Operators.sort(field=f"_id.{Record.INSTANTIATES_}", sort_order=1),
            Operators.group_by(group_key={Record.DATASET_: Record.DATASET___}, groups=[
                {"name": "orderedfeatures", "operator": "$push", "field": Record.INSTANTIATES___}
            ]),
            Operators.set_variables(variables=[
                {"name": "top10",
                 "operation": Operators.filter_array(input_array_name="$orderedfeatures", element="item", cond={},
                                                     limit=10)}
            ]),
            # Operators.limit(10),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                "top10": 1,
                "_id": 0
            }),
            Operators.unwind(field="top10"),
            Operators.project(field=None, projected_value={Record.INSTANTIATES_: "$top10", Record.DATASET_: 1}),
            Operators.write_to_table(table_name=TableNames.TOP10_FEATURES)
        ]
        log.info(top_10_features)
        database.db[TableNames.RECORD].aggregate(top_10_features)

        pairs_features = [
            Operators.cartesian_product(join_table_name=TableNames.TOP10_FEATURES, lookup_field_name="features_b",
                                        filter_dict={}),
            Operators.unwind(field="features_b"),
            Operators.match(field=None, value={"$expr": {
                "$and": [
                    {"$eq": [Record.DATASET__, f"$features_b.{Record.DATASET_}"]},
                    # created pairs of features in a single dataset
                    {"$lt": [Record.INSTANTIATES__, f"$features_b.{Record.INSTANTIATES_}"]}
                ]
            }}, is_regex=False),
            Operators.project(field=None,
                              projected_value={Record.DATASET_: Record.DATASET__, "feat_a": Record.INSTANTIATES__,
                                               "feat_b": f"$features_b.{Record.INSTANTIATES_}", "_id": 0}),
            # this allows to not export the _id in the table (which contains duplicated due to unwind), and _id is regenerated when creating the table
            Operators.write_to_table(TableNames.PAIRS_FEATURES)
        ]
        log.info(pairs_features)
        database.db[TableNames.TOP10_FEATURES].aggregate(pairs_features)

        # second part, compute the Pearson coefficient for each pair of features stored above within the collection
        return [
            Operators.lookup(
                join_table_name=TableNames.RECORD,
                local_field=None,
                foreign_field=None,
                let={"feat_a": "$feat_a", "a_dataset": Record.DATASET__},
                pipeline=[Operators.match(field=None, value={"$expr": {"$and": [
                    {"$eq": [Record.INSTANTIATES__, "$$feat_a"]},
                    {"$eq": [Record.DATASET__, "$$a_dataset"]}
                ]}}, is_regex=False)],
                lookup_field_name="x"
            ),
            Operators.lookup(
                join_table_name=TableNames.RECORD,
                local_field=None,
                foreign_field=None,
                let={"feat_b": "$feat_b", "b_dataset": Record.DATASET__},
                pipeline=[Operators.match(field=None, value={"$expr": {
                    "$and": [
                        {"$eq": [Record.INSTANTIATES__, "$$feat_b"]},
                        {"$eq": [Record.DATASET__, "$$b_dataset"]}
                    ]}}, is_regex=False)],
                lookup_field_name="y"
            ),
            Operators.project(field=None, projected_value={"values": {"$zip": {"inputs": ["$x", "$y"]}}}),
            Operators.unwind(field="values"),
            Operators.project(field=None, projected_value={"x": {"$arrayElemAt": ["$values", 0]},
                                                           "y": {"$arrayElemAt": ["$values", 1]}}),
            Operators.project(field=None, projected_value={"x": "$x.value", "y": f"$y.{Record.VALUE_}",
                                                           "feat_a": f"$x.{Record.INSTANTIATES_}",
                                                           "feat_b": f"$y.{Record.INSTANTIATES_}",
                                                           Record.DATASET_: f"$x.{Record.DATASET_}"}),
            Operators.group_by(group_key={"feat_a": "$feat_a", "feat_b": "$feat_b", Record.DATASET_: Record.DATASET__},
                               groups=[
                                   {"name": "originalX", "operator": "$push", "field": "$x"},
                                   {"name": "originalY", "operator": "$push", "field": "$y"},
                                   {"name": "avgX", "operator": "$avg", "field": "$x"},
                                   {"name": "avgY", "operator": "$avg", "field": "$y"},
                               ]),
            Operators.project(field=None, projected_value={
                # to subtract/map a number to each element of an array, the syntax is a bit more complex than "Xsubs": {"$subtract": ["$originalX", "$avgX"]}
                # it requires a map
                "subsX": {"$map": {"input": "$originalX", "as": "originalXval",
                                   "in": {"$subtract": ["$$originalXval", "$avgX"]}}},
                "subsY": {"$map": {"input": "$originalY", "as": "originalYval",
                                   "in": {"$subtract": ["$$originalYval", "$avgY"]}}}
            }),
            Operators.project(field=None, projected_value={
                "subsX": 1,
                "subsY": 1,
                "subsSquareX": {"$map": {"input": "$subsX", "as": "subsXVal", "in": {"$pow": ["$$subsXVal", 2]}}},
                "subsSquareY": {"$map": {"input": "$subsY", "as": "subsYVal", "in": {"$pow": ["$$subsYVal", 2]}}},
                "multXY": {"$map": {"input": {"$zip": {"inputs": ["$subsX", "$subsY"]}}, "as": "pair",
                                    "in": {
                                        "$multiply": [{"$arrayElemAt": ["$$pair", 0]}, {"$arrayElemAt": ["$$pair", 1]}]}
                                    }}
            }),
            Operators.project(field=None, projected_value={
                "sumXsubsSquare": {"$sum": "$subsSquareX"},
                # no need to group by in this case (and the group by does not work)
                "sumYsubsSquare": {"$sum": "$subsSquareY"},
                "sumXY": {"$sum": "$multXY"}
            }),
            Operators.project(field=None, projected_value={
                "pearson": {
                    "$cond": {
                        "if": {"$or": [{"$eq": ["$sumXsubsSquare", 0]}, {"$eq": ["$sumYsubsSquare", 0]}]},
                        "then": 0,
                        "else": {
                            "$divide": ["$sumXY", {"$sqrt": {"$multiply": ["$sumXsubsSquare", "$sumYsubsSquare"]}}]
                        }
                    }
                }
            }),
            # the last group by and project groups all features b within the feature a for easier profiles
            Operators.group_by(group_key={Record.INSTANTIATES_: "$_id.feat_a", Record.DATASET_: Record.DATASET___},
                               groups=[
                                   {"name": "feat_b_dict", "operator": "$push",
                                    "field": {"k": "$_id.feat_b", "v": "$pearson"}}
                               ]),
            Operators.project(field=None, projected_value={
                Record.INSTANTIATES_: 1,
                "pearson": {
                    "$arrayToObject": {
                        "$map": {
                            "input": "$feat_b_dict",
                            "as": "item",
                            "in": [{"$toString": "$$item.k"}, {"coefficient": "$$item.v"}]
                        }
                    }
                }
            })
        ]

    def imbalance_query(self, features_ids: list) -> list:
        # Ratio between the number of appearances of the most frequent value and the least frequent value.
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1}
                ]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "max_freq", "operator": "$max", "field": "$frequency"},
                    {"name": "min_freq", "operator": "$min", "field": "$frequency"},
                ]),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "imbalance": {
                    "$cond": {
                        "if": {"$eq": ["$min_freq", 0]},
                        "then": 0,
                        "else": {"$divide": ["$max_freq", "$min_freq"]}
                    }
                }
            })
        ]

    def constancy_query(self, features_ids: list) -> list:
        # Ratio between the number of appearances of the most frequent value and the number of non-null values.
        # However, we never have null values because we do not create Records for them
        # thus, the constancy is always 1 (for practical reasons, I divided the max freq by itself, to always obtain 1)
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1},
                    {"name": "frequencyNonNull", "operator": "$sum", "field": 1}
                ]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "max_freq", "operator": "$max", "field": "$frequency"}
                ]),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "constancy": {
                    "$cond": {
                        "if": {"$eq": ["$max_freq", 0]},
                        "then": 0,
                        "else": {"$divide": ["$max_freq", "$max_freq"]}
                    }
                }
            })
        ]

    def mode_query(self, features_ids: list) -> list:
        # the most frequent value
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1},
                ]),
            Operators.sort(field=f"_id.{Record.INSTANTIATES_}", sort_order=1),
            Operators.sort(field="frequency", sort_order=-1),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "values", "operator": "$push", "field": Record.VALUE___}
                ]),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "mode": {"$arrayElemAt": ["$values", 0]}
            })
        ]

    def uniqueness_query(self, features_ids: list) -> list:
        # Percentage of distinct values with respect to the total amount of non-null values
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1}]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "count", "operator": "$sum", "field": "$frequency"},
                    {"name": "distinct_count", "operator": "$sum", "field": 1}
                ]
                ),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "uniqueness": {
                    "$cond": {
                        "if": {"$eq": ["$count", 0]},
                        "then": 0,
                        "else": {
                            "$divide": ["$distinct_count", "$count"]
                        }
                    }
                }
            })
        ]

    def entropy_query(self, features_ids: list) -> list:
        # Measure of uncertainty and disorder within the values of the column.
        # A large entropy means that the values are highly heterogeneous.
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1}]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "total", "operator": "$sum", "field": "$frequency"},
                    {"name": "frequencies", "operator": "$push", "field": "$frequency"}
                ]),
            Operators.unwind("frequencies"),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "prob": {
                    "$cond": {
                        "if": {"$eq": ["$total", 0]},
                        "then": [0],
                        "else": {"$divide": ["$frequencies", "$total"]}
                    }
                }
            }),
            Operators.project(field=None, projected_value={
                "_id": "$_id",
                "entropy_value": {"$multiply": ["$prob", {"$log": ["$prob", 2]}]}
            }),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "entropy", "operator": "$sum", "field": "$entropy_value"}
                ]),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "entropy": {"$abs": "$entropy"}
            })
        ]

    def density_query(self, features_ids: list) -> list:
        # a measure of appropriate numerosity and intensity between different real-world entities available in the data
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__}, groups=[
                    {"name": "frequency", "operator": "$sum", "field": 1}]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "total", "operator": "$sum", "field": "$frequency"},
                    {"name": "distinct_count", "operator": "$sum", "field": 1},
                    {"name": "frequencies", "operator": "$push", "field": "$frequency"}
                ]),
            Operators.unwind("frequencies"),
            Operators.project(field=None, projected_value={
                "dataset": Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "prob": {
                    "$cond": {
                        "if": {"$eq": ["$total", 0]},
                        "then": [0],
                        "else": {"$divide": ["$frequencies", "$total"]}
                    }
                },
                "distinct_count": "$distinct_count"
            }),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "avg_dens_value", "operator": "$avg", "field": "$prob"},
                    {"name": "probs", "operator": "$push", "field": "$prob"},
                    {"name": "distinct_count", "operator": "$first", "field": "$distinct_count"}
                ]),
            Operators.unwind("probs"),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "density_value": {"$abs": {"$subtract": ["$probs", "$avg_dens_value"]}},
                "distinct_count": "$distinct_count"
            }),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    {"name": "densities_sum", "operator": "$sum", "field": "$density_value"},
                    {"name": "distinct_count", "operator": "$first", "field": "$distinct_count"}
                ]),
            Operators.project(field=None, projected_value={
                Record.DATASET_: Record.DATASET___,
                Record.INSTANTIATES_: Record.INSTANTIATES___,
                "density": {
                    "$cond": {
                        "if": {"$eq": ["$distinct_count", 0]},
                        "then": 0,
                        "else": {"$divide": ["$densities_sum", "$distinct_count"]}
                    }
                }
            })
        ]

    def values_and_counts_query(self, features_ids: list) -> list:
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           Record.VALUE_: Record.VALUE__},
                groups=[{"name": "counts", "operator": "$sum", "field": 1}]),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES___, Record.DATASET_: Record.DATASET___}, groups=[
                    # push in an array of maps of two elements (value and counts) the elements of the  pair (dataset, feature)
                    {"name": "values_and_counts", "operator": "$push", "field": {Record.VALUE_: {
                        "$cond": {"if": {"$eq": [Record.VALUE___, DEFAULT_NAN_VALUE]}, "then": None,
                                  "else": Record.VALUE___}}, "count": "$counts"}}
                ])
        ]

    def data_type_validity_query(self, features_ids: list) -> list:
        reverse_mapping = {"int": "int", "double": "float", "string": "str"}
        type_switch = {
            "$switch": {
                "branches": [
                    {
                        "case": {"$eq": [{"$type": "$value"}, bson_type]},
                        "then": user_alias
                    }
                    for bson_type, user_alias in reverse_mapping.items()
                ],
                "default": "unknown"
            }
        }

        # db["Record"].aggregate([
        # { "$addFields": { "featureType": { "$type": "$value" } } },
        # { "$group": { "_id": { "instantiates": "$instantiates", "dataset": "$dataset", "featureType": "$featureType" }, "count": { "$sum": 1 } } },
        # { "$set": {"featureTypeAlias":{"$switch":{"branches":[{"case":{"$eq":["$_id.featureType","int"]},"then":"int"},{"case":{"$eq":["$_id.featureType","double"]},"then":"float"},{"case":{"$eq":["$_id.featureType","string"]},"then":"str"},{"case":{"$eq":["$_id.featureType","date"]},"then":"date"},{"case":{"$eq":["$_id.featureType","timestamp"]},"then":"datetime"},{"case":{"$eq":["$_id.featureType","bool"]},"then":"bool"},{"case":{"$eq":["$_id.featureType","object"]},"then":"category"}],"default":"unknown"}}}}
        # { "$lookup": { "from": "Feature", "localField": "_id.instantiates", "foreignField": "identifier", "let": { "featureTypess": "$featureTypeAlias" }, "pipeline": [{ "$match": { "$expr": { "$eq": ["$data_type", "$$featureTypess"] } } }], "as": "common_fid" } },
        # { "$lookup": { "from": "TMP_CountsFeatures", "localField": "_id.instantiates", "foreignField": "identifier", "as": "counts_fid" } },
        # { "$set": { "f_all_counts": { "$arrayElemAt": ["$counts_fid.count_all_values", 0]}}},
        # { "$project": {"instantiates": "$_id.instantiates", "dt_validity": {"$divide": ["$count", "$f_all_counts"]}}}
        # ] )
        return [
            Operators.match(field=Record.INSTANTIATES_, value={"$in": features_ids}, is_regex=False),
            Operators.add_fields(key="featureType", value={"$type": Record.VALUE__}),
            Operators.group_by(
                group_key={Record.INSTANTIATES_: Record.INSTANTIATES__, Record.DATASET_: Record.DATASET__,
                           "featureType": "$featureType"},
                groups=[{"name": "count", "operator": "$sum", "field": 1}]),
            Operators.set_variables(variables=[{"name": "featureTypeAlias", "operation": type_switch}]),
            Operators.lookup(join_table_name=TableNames.FEATURE, local_field=f"_id.{Record.INSTANTIATES_}",
                             foreign_field=Feature.IDENTIFIER_, let={"featureTypess": "$featureTypeAlias"},
                             pipeline=[{"$match": {"$expr": {"$eq": [Feature.DT__, "$$featureTypess"]}}}],
                             lookup_field_name="common_fid"),
            Operators.lookup(join_table_name=TableNames.COUNTS_FEATURES, local_field=f"_id.{Record.INSTANTIATES_}", foreign_field=Feature.IDENTIFIER_, lookup_field_name="counts_fid", let=None, pipeline=None),
            Operators.set_variables(variables=[{"name": "f_all_counts", "operation": {"$arrayElemAt": ["$counts_fid.count_all_values", 0]}}]),
            Operators.project(field={"data_type_validity": {"$divide": ["$count", "$f_all_counts"]}}, projected_value=None)
        ]

    def missing_percentage_query_without_samples(self, features_ids: list) -> list:
        return [
            # first compute the missing percentage of each feature by dividing the number of values and the number of patients
            Operators.match(field=None, value={"$and": [{Record.INSTANTIATES_: {"$in": features_ids}}, {
                Record.ENTITY_TYPE_: {"$ne": f"{Profile.CLINICAL}{TableNames.RECORD}"}}]}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "count_db_values", "operator": "$sum", "field": 1}
                ]),
            Operators.lookup(join_table_name=TableNames.COUNTS_PATIENTS, foreign_field=f"_id.{Record.DATASET_}",
                             local_field=f"_id.{Record.DATASET_}", lookup_field_name="joined", let=None, pipeline=None),
            Operators.project(field=None, projected_value={
                "missing_percentage": {
                    "$cond": {
                        "if": {"$eq": [{"$arrayElemAt": ["$joined.distinct_nb_patients", 0]}, 0]},
                        "then": 0,
                        "else": {"$subtract": [1, {
                            "$divide": ["$count_db_values", {"$arrayElemAt": ["$joined.distinct_nb_patients", 0]}]}]}
                    }
                }
            })
        ]

    def missing_percentage_query_with_samples(self, features_ids: list) -> list:
        return [
            # first compute the missing percentage of each feature by dividing the number of values and the number of patients
            Operators.match(field=None, value={"$and": [{Record.INSTANTIATES_: {"$in": features_ids}}, {
                Record.ENTITY_TYPE_: {"$eq": f"{Profile.CLINICAL}{TableNames.RECORD}"}}]}, is_regex=False),
            Operators.group_by(
                group_key={Record.DATASET_: Record.DATASET__, Record.INSTANTIATES_: Record.INSTANTIATES__}, groups=[
                    {"name": "count_db_values", "operator": "$sum", "field": 1}
                ]),
            Operators.lookup(join_table_name=TableNames.COUNTS_SAMPLES, foreign_field=Record.DATASET_,
                             local_field=Record.DATASET_, lookup_field_name="joined", let=None, pipeline=None),
            Operators.project(field=None, projected_value={
                "missing_percentage": {
                    "$cond": {
                        "if": {"$eq": [{"$arrayElemAt": ["$joined.distinct_nb_samples", 0]}, 0]},
                        "then": 0,
                        "else": {"$subtract": [1, {
                            "$divide": ["$count_db_values", {"$arrayElemAt": ["$joined.distinct_nb_samples", 0]}]}]}
                    }
                }
            })
        ]

    def finalize_query(self, include_value: bool) -> list:
        # get dataset and instantiates from the group key and keep remaining fields, except the group key _id
        # project requires to specify all fields to keep, set and unset work only on those of interest and keep others
        if include_value:
            last_fields = [{"$set": {Record.DATASET_: Record.DATASET___, Record.INSTANTIATES_: Record.INSTANTIATES___,
                                     Record.VALUE_: Record.VALUE___}}]
        else:
            last_fields = [{"$set": {Record.DATASET_: Record.DATASET___, Record.INSTANTIATES_: Record.INSTANTIATES___}}]
        last_fields.append({"$unset": ["_id"]})
        # merge the profile into the profile table
        last_fields.append(
            Operators.merge(table_name=TableNames.FEATURE_PROFILE,
                            on_attribute=[Record.DATASET_, Record.INSTANTIATES_],
                            when_matched="merge",
                            when_not_matched="insert"))
        return last_fields

    def to_json(self):
        return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



import copy
import json
import os.path
import random
import unittest

import pytest
from jsonlines import jsonlines

from constants.defaults import NO_ID
from constants.structure import TEST_DB_NAME
from database.Counter import Counter
from database.Database import Database
from database.Execution import Execution
from entities.Record import Record
from entities.Resource import Resource
from entities.ResourceTest import ResourceTest
from enums.HospitalNames import HospitalNames
from enums.ParameterKeys import ParameterKeys
from enums.TableNames import TableNames
from utils.file_utils import write_in_file, from_json_line_to_json_str, get_json_resource_file
from utils.test_utils import wrong_number_of_docs, compare_tuples, set_env_variables_from_dict


class TestDatabase(unittest.TestCase):
    execution = Execution()

    def setUp(self):
        # before each test, get back to the original test configuration
        args = {
            ParameterKeys.DB_NAME: TEST_DB_NAME,
            ParameterKeys.DB_DROP: "True",
            ParameterKeys.HOSPITAL_NAME: HospitalNames.TEST_H1
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)

    def test_check_server_is_up(self):
        # test with the correct (default) string
        _ = Database(execution=TestDatabase.execution)  # this should return no exception (successful connection)
        # database.close()

        # test with a wrong connection string
        # NB Aug 20, 2024: this cannot be tested anymore because the MongoDB uri is now internal to Docker, thus cannot be changed.
        # set_env_variables_from_dict(env_vars={ ExecutionKeys.DB_CONNECTION_KEY: "a_random_string" })
        # log.debug("Set up in test")
        # TestDatabase.execution.set_up(setup_data_files=False)
        # with pytest.raises(ConnectionError):
        #     _ = Database(execution=TestDatabase.execution)  # this should return an exception (broken connection) because check_server_is_up() will return one

    def test_drop(self):
        # create a test database
        # and add only one triple to be sure that the db is created
        database = Database(execution=TestDatabase.execution)
        database.db[TableNames.TEST].insert_one(document={"id": 1, "name": "Alice Doe"})
        assert database.db_exists(TEST_DB_NAME) is True, "The database does not exist."
        database.drop_db()
        # check the DB does not exist anymore after drop
        assert database.db_exists(TEST_DB_NAME) is False, "The database has not been dropped."

    def test_no_drop(self):
        # test that the database is not dropped when the user asks for it
        # 1. create a db with a single table and only one tuple
        database = Database(execution=TestDatabase.execution)
        database.db[TableNames.TEST].insert_one(document={"id": 1, "name": "Alice Doe"})
        assert database.db_exists(TEST_DB_NAME) is True, "The database does not exist."  # we make sure that the db exists

        # 2. we create a new instance to the same database, with drop_db=False
        set_env_variables_from_dict(env_vars={ParameterKeys.DB_DROP: "False"})
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)

        # 3. check that the database still exists (i.e., the constructor did not reset it)
        assert database.db_exists(TEST_DB_NAME) is True, "The database does not exist."

        # 4. check that the database still exists, even after a drop
        database.drop_db()  # this should do nothing as we explicitly specified to not drop the db
        assert database.db_exists(TEST_DB_NAME) is True, "The database has not been dropped."

    def test_insert_empty_tuple(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {}
        my_original_tuple = copy.deepcopy(my_tuple)
        database.insert_one_tuple(table_name=TableNames.TEST, one_tuple=my_tuple)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])

    def test_insert_one_tuple(self):
        # 1. create a db with a single table and only one tuple
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"id": "1", "name": "Alice Doe"}
        my_original_tuple = copy.deepcopy(my_tuple)
        database.insert_one_tuple(table_name=TableNames.TEST, one_tuple=my_tuple)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]

        assert len(docs) == 1, wrong_number_of_docs(1)
        # if assert does not fail, we indeed have one tuple, thus we can check its attributes directly
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])

    def test_insert_no_tuples(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = []

        with pytest.raises(TypeError):
            # mongoDB does not allow to insert an empty array
            database.insert_many_tuples(table_name=TableNames.TEST, tuples=my_tuples)

    def test_insert_many_tuples(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [{"id": 1, "name": "Louise", "country": "FR", "job": "PhD student"},
                      {"id": 2, "name": "Francesca", "country": "IT", "university": True},
                      {"id": 3, "name": "Martin", "country": "DE", "age": 26}]
        my_original_tuples = copy.deepcopy(my_tuples)  # we need to do a deep copy because MongoDB's insert does modify the given list of tuples
        database.insert_many_tuples(table_name=TableNames.TEST, tuples=my_tuples)
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"id": 1})]

        assert len(my_original_tuples) == len(docs), wrong_number_of_docs(len(my_original_tuples))
        # the three tuples are sorted, so we can iterate over them easily
        for i in range(len(my_original_tuples)):
            compare_tuples(original_tuple=my_original_tuples[i], inserted_tuple=docs[i])

    def test_upsert_one_tuple_single_key(self):
        # remark: even if that would be cleaner, I cannot separate each individual upsert test
        # because we don't know in advance the test execution order
        # and, they need to be executed in order to test properly whether upsert works
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"name": "Nelly", "age": 26}
        my_original_tuple = copy.deepcopy(my_tuple)
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name"], one_tuple=my_tuple)

        # 1. first, we check that the initial upserted tuple has been correctly inserted
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])

        # 2. We upsert the exact same tuple:
        # there should be no duplicate
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name"], one_tuple=my_tuple)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])  # we also check that the tuple did not change

        # 3. We upsert the same tuple with a different age with REPLACE:
        # no new tuple should be inserted (that document already exists)
        # the tuple should be updated
        args = {
            ParameterKeys.DB_DROP: "False"
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)
        my_tuple_age = {"name": "Nelly", "age": 27, "city": "Lyon"}  # same as my_tuple but with a different age and a new field city
        my_original_tuple_age = copy.deepcopy(my_tuple_age)
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name"], one_tuple=my_tuple_age)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple_age, inserted_tuple=docs[0])  # we check that the tuple did change (REPLACE)

    def test_upsert_one_tuple_unknown_key(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"name": "Nelly", "age": 26}

        with pytest.raises(KeyError):
            # this should return an exception (value error) because the upsert contains an unknown key (city does not exist in my_tuple)
            database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["city"], one_tuple=my_tuple)

    def test_upsert_one_tuple_multi_key(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"name": "Nelly", "age": 26}
        my_original_tuple = copy.deepcopy(my_tuple)
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name", "age"], one_tuple=my_tuple)

        # 1. first, we check that the initial upserted tuple has been correctly inserted
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])

        # 2. We upsert the exact same tuple:
        # there should be no duplicate
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name", "age"], one_tuple=my_tuple)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 1, wrong_number_of_docs(1)
        compare_tuples(original_tuple=my_original_tuple, inserted_tuple=docs[0])  # we also check that the tuple did not change

        # 3. We upsert the same tuple with a different age with REPLACE:
        # a new tuple is added because no existing tuple has this combination (name, age)
        # and the current one should not be updated because no tuple should have matched
        args = {
            ParameterKeys.DB_DROP: "False"
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)
        my_tuple_age = {"name": "Nelly", "age": 27, "city": "Lyon"}  # same as my_tuple but with a different age and a new field city
        my_original_tuple_age = copy.deepcopy(my_tuple_age)
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name", "age"], one_tuple=my_tuple_age)
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1, "age": 1})]
        expected_docs = [my_original_tuple, my_original_tuple_age]
        assert len(docs) == 2, wrong_number_of_docs(2)
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 4. We upsert a similar tuple (same name and age but different city) with REPLACE:
        # no new tuple should be inserted (that document does already exist)
        # the former one should have changed (because REPLACE)
        args = {
            ParameterKeys.DB_DROP: "False"
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)
        my_new_tuple = {"name": "Nelly", "age": 26, "city": "Lyon"}
        my_original_new_tuple = copy.deepcopy(my_new_tuple)
        database.upsert_one_tuple(table_name=TableNames.TEST, unique_variables=["name", "age"],
                                  one_tuple=my_new_tuple)
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1, "age": 1})]
        expected_docs = [my_original_new_tuple, my_original_tuple_age]  # alphabetical order
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

    def test_upsert_one_batch_of_tuples_single_key(self):
        database = Database(execution=TestDatabase.execution)
        # 1. upsert an initial batch of 2 tuples
        my_batch = [{"name": "Nelly", "age": 26}, {"name": "Julien", "age": 30, "city": "Lyon"}]
        my_original_batch = copy.deepcopy(my_batch)
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name"], the_batch=my_batch, ordered=True)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == 2, wrong_number_of_docs(2)
        for i in range(len(my_original_batch)):
            compare_tuples(original_tuple=my_original_batch[i], inserted_tuple=docs[i])

        # 2. upsert a batch with one similar tuple, and one different
        my_batch_2 = [{"name": "Nelly", "age": 27}, {"name": "Pietro", "city": "Milano"}]
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name"], the_batch=my_batch_2, ordered=True)
        expected_docs = [my_original_batch[1], my_batch_2[0], my_batch_2[1]]  # ordered by name
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1})]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 3. upsert a batch with one similar tuple and one different
        # with upsert policy being REPLACE
        args = {
            ParameterKeys.DB_DROP: "False"
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)
        my_batch_3 = [{"name": "Nelly", "age": 27}, {"name": "Anna", "citizenship": "Italian"}]
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name"], the_batch=my_batch_3, ordered=True)
        expected_docs = [my_batch_3[1], my_original_batch[1], my_batch_3[0], my_batch_2[1]]  # ordered by name
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1})]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

    def test_upsert_one_batch_of_tuples_multi_key(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [{"name": "Nelly", "age": 26}, {"name": "Julien", "age": 30}]
        my_original_tuples = copy.deepcopy(my_tuples)
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name", "age"], the_batch=my_tuples, ordered=True)

        # 1. first, we check that the initial upserted tuple has been correctly inserted
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == len(my_original_tuples), wrong_number_of_docs(len(my_original_tuples))
        for i in range(len(my_original_tuples)):
            compare_tuples(original_tuple=my_original_tuples[i], inserted_tuple=docs[i])

        # 2. We upsert the exact same batch of tuples:
        # there should be no duplicate
        my_original_tuples_2 = copy.deepcopy(my_original_tuples)
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name", "age"], the_batch=my_original_tuples, ordered=True)
        docs = [doc for doc in database.db[TableNames.TEST].find({})]
        assert len(docs) == len(my_original_tuples_2), wrong_number_of_docs(len(my_original_tuples_2))
        for i in range(len(my_original_tuples_2)):
            compare_tuples(original_tuple=my_original_tuples_2[i], inserted_tuple=docs[i])

        # 3. We upsert one (out of 2) same tuple with a different city with REPLACE:
        # a new tuple is added because no existing tuple has this combination (name, age)
        # and the current ones should not be updated because no tuple should have matched
        args = {
            ParameterKeys.DB_DROP: "False"
        }
        set_env_variables_from_dict(env_vars=args)
        TestDatabase.execution.internals_set_up()
        TestDatabase.execution.file_set_up(setup_files=False)
        database = Database(execution=TestDatabase.execution)
        my_tuples_age_2 = [{"name": "Nelly", "age": 26, "city": "Paris"}, {"name": "Pietro", "age": -1, "city": "Milano"}]
        my_original_tuple_age_2 = copy.deepcopy(my_tuples_age_2)
        database.upsert_one_batch_of_tuples(table_name=TableNames.TEST, unique_variables=["name", "age"], the_batch=my_tuples_age_2, ordered=True)
        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1, "age": 1})]
        expected_docs = [my_original_tuples[1], my_original_tuple_age_2[0], my_original_tuple_age_2[1]]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

    def test_retrieve_resource_identifiers_1(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"identifier": 123, "name": "Nelly"}
        database.db[TableNames.TEST].insert_one(my_tuple)
        the_doc = database.retrieve_mapping(table_name=TableNames.TEST, key_fields="name", value_fields="identifier", filter_dict={})
        expected_doc = {"Nelly": 123}
        assert the_doc == expected_doc

    def test_retrieve_resource_identifiers_10(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [{Resource.IDENTIFIER_: i, Record.VALUE_: i + random.randint(0, 100)} for i in range(0, 10)]
        my_original_tuples = copy.deepcopy(my_tuples)
        database.db[TableNames.TEST].insert_many(my_tuples)
        docs = database.retrieve_mapping(table_name=TableNames.TEST, key_fields=Record.VALUE_, value_fields=Resource.IDENTIFIER_, filter_dict={})
        expected_docs = {}
        for doc in my_original_tuples:
            expected_docs[doc[Record.VALUE_]] = doc[Resource.IDENTIFIER_]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for key, value in expected_docs.items():
            compare_tuples(original_tuple={key: value}, inserted_tuple={key: docs[key]})

    def test_retrieve_resource_identifiers_wrong_key(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"identifier": 123, "name": "Nelly"}
        database.db[TableNames.TEST].insert_one(my_tuple)
        with pytest.raises(KeyError):
            _ = database.retrieve_mapping(table_name=TableNames.TEST, key_fields="name2", value_fields="identifier", filter_dict={})

    def test_retrieve_patient_identifiers_1(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"identifier": 123, "name": "Nelly"}
        database.db[TableNames.TEST].insert_one(my_tuple)
        the_doc = database.retrieve_mapping(table_name=TableNames.TEST, key_fields="name", value_fields="identifier", filter_dict={})
        expected_doc = {"Nelly": 123}
        assert the_doc == expected_doc

    def test_retrieve_patient_identifiers_10(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [{Resource.IDENTIFIER_: i, Record.VALUE_: i + random.randint(0, 100)} for i in range(0, 10)]
        my_original_tuples = copy.deepcopy(my_tuples)
        database.db[TableNames.TEST].insert_many(my_tuples)
        docs = database.retrieve_mapping(table_name=TableNames.TEST, key_fields=Record.VALUE_, value_fields=Resource.IDENTIFIER_, filter_dict={})
        expected_docs = {}
        for doc in my_original_tuples:
            expected_docs[doc[Record.VALUE_]] = doc[Resource.IDENTIFIER_]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for key, value in expected_docs.items():
            compare_tuples(original_tuple={key: value}, inserted_tuple={key: docs[key]})

    def test_retrieve_patient_identifiers_wrong_key(self):
        database = Database(execution=TestDatabase.execution)
        my_tuple = {"identifier": 123, "name": "Nelly"}
        database.db[TableNames.TEST].insert_one(my_tuple)
        with pytest.raises(KeyError):
            _ = database.retrieve_mapping(table_name=TableNames.TEST, key_fields="name2", value_fields="identifier", filter_dict={})

    def test_write_in_file(self):
        counter = Counter()
        my_tuples = [
            ResourceTest(identifier=NO_ID, counter=counter),
            ResourceTest(identifier=NO_ID, counter=counter),
            ResourceTest(identifier=NO_ID, counter=counter),
        ]
        my_tuples_as_json = [my_tuples[i].to_json() for i in range(len(my_tuples))]

        write_in_file(resource_list=my_tuples_as_json, current_working_dir=self.execution.working_dir_current, table_name=TableNames.TEST, is_feature=False, dataset_id=1, to_json=False)
        filepath = get_json_resource_file(current_working_dir=self.execution.working_dir_current, table_name=TableNames.TEST, dataset_id=1)
        assert os.path.exists(filepath) is True
        with jsonlines.open(filepath) as my_file:
            read_tuples = [obj for obj in my_file]
            assert len(my_tuples_as_json) == len(read_tuples), wrong_number_of_docs(len(my_tuples))
            assert my_tuples_as_json == read_tuples

    def test_write_in_file_no_resource(self):
        _ = Database(execution=TestDatabase.execution)
        my_tuples = []
        write_in_file(resource_list=my_tuples, current_working_dir=self.execution.working_dir_current, table_name=TableNames.TEST, is_feature=False, dataset_id=98, to_json=False)
        filepath = get_json_resource_file(current_working_dir=self.execution.working_dir_current, table_name=TableNames.TEST, dataset_id=98)
        assert os.path.exists(filepath) is False  # no file should have been created since there is no data to write

    def test_load_json_in_table(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"name": "Nelly", "age": 26},
            {"name": "Julien", "age": 30, "city": "Lyon"},
            {"name": "Julien", "age": 30, "city": "Paris"},
            {"name": "Nelly", "age": 27, "job": "post-doc"},
            {"name": "Pietro", "age": -1, "country": "Italy"}
        ]
        my_original_tuples = copy.deepcopy(my_tuples)

        # I need to write the tuples in the working dir
        # because load_json_in_table() looks for files having the given table name in that directory
        write_in_file(resource_list=my_tuples, current_working_dir=self.execution.working_dir_current, table_name=TableNames.TEST, is_feature=False, dataset_id=97, to_json=False)
        database.load_json_in_table_for_tests(unique_variables=["name", "age"], dataset_id=97)

        docs = [doc for doc in database.db[TableNames.TEST].find({}).sort({"name": 1, "age": 1})]
        expected_docs = [my_original_tuples[2], my_original_tuples[0], my_original_tuples[3], my_original_tuples[4]]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

    def test_find_operation_1(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"name": "Julien", "job": "engineer", "city": "Lyon"},
            {"name": "Nelly", "job": "bachelor student", "city": "Lyon"},
            {"name": "Nelly", "job": "master student", "city": "Lyon"},
            {"name": "Nelly", "job": "phd", "city": "Massy"},
            {"name": "Nelly", "job": "post-doc", "city": "Milano"},
            {"name": "Pietro", "job": "assistant prof.", "citizenship": "Italian"}
        ]
        my_original_tuples = copy.deepcopy(my_tuples)
        database.db[TableNames.TEST].insert_many(documents=my_tuples)

        # 1. no filter, no projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={}, projection={}).sort({"name": 1})]
        assert len(docs) == len(my_original_tuples), wrong_number_of_docs(len(my_original_tuples))
        for i in range(len(my_original_tuples)):
            compare_tuples(original_tuple=my_original_tuples[i], inserted_tuple=docs[i])

        # 2. a single filter, no projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"city": "Lyon"}, projection={})]
        expected_docs = [my_original_tuples[0], my_original_tuples[1], my_original_tuples[2]]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 3. a multi filter, no projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"name": "Nelly", "city": "Lyon"}, projection={})]
        expected_docs = [my_original_tuples[1], my_original_tuples[2]]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 4. no filter, a single projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={}, projection={"name": 1}).sort({"name": 1})]
        expected_docs = [{"name": doc["name"]} for doc in my_original_tuples]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 5. a single filter, a single projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"city": "Lyon"}, projection={"name": 1})]
        expected_docs = [{"name": my_original_tuples[0]["name"]}, {"name": my_original_tuples[1]["name"]}, {"name": my_original_tuples[2]["name"]}]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 6. a multi filter, a single projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"name": "Nelly", "city": "Lyon"}, projection={"name": 1})]
        expected_docs = [{"name": my_original_tuples[1]["name"]}, {"name": my_original_tuples[2]["name"]}]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 7. no filter, a multi projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={}, projection={"name": 1, "job": 1}).sort({"name": 1, "job": 1})]
        expected_docs = [{"name": doc["name"], "job": doc["job"]} for doc in my_original_tuples]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 8. a single filter, a multi projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"city": "Lyon"}, projection={"name": 1, "job": 1})]
        expected_docs = [
            {"name": my_original_tuples[0]["name"], "job": my_original_tuples[0]["job"]},
            {"name": my_original_tuples[3]["name"], "job": my_original_tuples[1]["job"]},
            {"name": my_original_tuples[4]["name"], "job": my_original_tuples[2]["job"]}
        ]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

        # 9. a multi filter, a multi projection
        docs = [doc for doc in database.find_operation(table_name=TableNames.TEST, filter_dict={"name": "Nelly", "city": "Lyon"}, projection={"name": 1, "job": 1})]
        expected_docs = [
            {"name": my_original_tuples[3]["name"], "job": my_original_tuples[1]["job"]},
            {"name": my_original_tuples[4]["name"], "job": my_original_tuples[2]["job"]}
        ]
        assert len(docs) == len(expected_docs), wrong_number_of_docs(len(expected_docs))
        for i in range(len(expected_docs)):
            compare_tuples(original_tuple=expected_docs[i], inserted_tuple=docs[i])

    def test_count_documents(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"name": "Julien", "from": "France"},
            {"name": "Nelly", "from": "France", "city": "Milano"},
            {"name": "Anna", "from": "Italy", "city": "Milano"}
        ]
        my_original_tuples = copy.deepcopy(my_tuples)
        database.db[TableNames.TEST].insert_many(documents=my_tuples)

        # 1. count with no filter
        current_count = database.count_documents(table_name=TableNames.TEST, filter_dict={})
        assert current_count == len(my_original_tuples), wrong_number_of_docs(len(my_original_tuples))

        # 2. count with single filter
        expected_docs = [my_original_tuples[0], my_original_tuples[1]]
        current_count = database.count_documents(table_name=TableNames.TEST, filter_dict={"from": "France"})
        assert current_count == len(expected_docs), wrong_number_of_docs(len(expected_docs))

        # 3. count with multi filter
        expected_docs = [my_original_tuples[1]]
        current_count = database.count_documents(table_name=TableNames.TEST, filter_dict={"from": "France", "name": "Nelly"})
        assert current_count == len(expected_docs), wrong_number_of_docs(len(expected_docs))

    def test_get_min_or_max_value(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"value": 0.2},
            {"value": 1},
            {"value": 10},
            {"value": 2}
        ]
        database.db[TableNames.TEST].insert_many(documents=my_tuples)
        min_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=1)
        max_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=-1)
        assert min_value == 0.2, "The expected minimum value is 0.2."
        assert max_value == 10, "The expected maximum value is 10."

        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"value": -0.2},
            {"value": -1},
            {"value": -10},
            {"value": -2}
        ]
        database.db[TableNames.TEST].insert_many(documents=my_tuples)
        min_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=1)
        max_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=-1)
        assert min_value == -10, "The expected minimum value is -10."
        assert max_value == -0.2, "The expected maximum value is -0.2."

    def test_get_min_or_max_resource_id(self):
        database = Database(execution=TestDatabase.execution)
        my_tuples = [
            {"value": 45},
            {"value": 54},
            {"value": 9},
            {"value": 154}
        ]
        database.db[TableNames.TEST].insert_many(documents=my_tuples)
        min_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=1)
        max_value = database.get_min_or_max_value(table_name=TableNames.TEST, field="value", sort_order=-1)
        assert min_value == 9, "The expected minimum value is 9."
        assert max_value == 154, "The expected maximum value is 154."

    def test_get_max_resource_counter_id(self):
        database = Database(execution=TestDatabase.execution)
        my_resources_1 = [
            {"identifier": 1, "name": "Anna"},
            {"identifier": 4, "name": "Julien"},
            {"identifier": 999, "name": "Nelly"},
        ]
        my_resources_2 = [
            {"identifier": 2, "name": "Anna"},
            {"identifier": 100, "name": "Nelly"},
            {"identifier": 998, "name": "Pietro"},
        ]
        # as an exception, we insert into LABORATORY_RECORD, not in TableNames.TEST,
        # because the method is made to set up resource counter and is expected to work on the
        # TableNames table names only
        database.db[TableNames.RECORD].insert_many(documents=my_resources_1)
        database.db[TableNames.FEATURE].insert_many(documents=my_resources_2)
        max_resource_id = database.get_max_resource_counter_id()
        assert max_resource_id == 999, "The expected max resource id is 999."



import locale
import os

import pandas as pd

from catalogue.FeatureProfileComputation import FeatureProfileComputation
from constants.defaults import NO_ID
from constants.structure import DOCKER_FOLDER_DATA
from database.Counter import Counter
from database.Database import Database
from entities.Dataset import Dataset
from database.Execution import Execution
from entities.Hospital import Hospital
from enums.MetadataColumns import MetadataColumns
from enums.Profile import Profile
from enums.TableNames import TableNames
from enums.TimerKeys import TimerKeys
from etl.Extract import Extract
from etl.Load import Load
from etl.Reporting import Reporting
from etl.Transform import Transform
from main_statistics.DatabaseStatistics import DatabaseStatistics
from main_statistics.QualityStatistics import QualityStatistics
from main_statistics.TimeStatistics import TimeStatistics
from utils.file_utils import read_tabular_file_as_string, write_in_file
from utils.setup_logger import log


class ETL:
    def __init__(self, execution: Execution, database: Database):
        self.execution = execution
        self.database = database
        self.datasets = []

        # set the locale
        log.debug(f"use locale: {self.execution.use_locale}")
        locale.setlocale(category=locale.LC_NUMERIC, locale=self.execution.use_locale)
        log.info(f"Current locale is: {locale.getlocale(locale.LC_NUMERIC)}")

        # init ETL steps
        self.extract = None
        self.transform = None
        self.load = None
        self.reporting = None
        self.profile_computation = None

    def run(self) -> None:
        time_stats = TimeStatistics(record_stats=True)
        time_stats.start(dataset=None, key=TimerKeys.TOTAL_TIME)
        compute_indexes = False

        quality_stats = QualityStatistics(record_stats=True)
        all_filenames = os.getenv("DATA_FILES").split(",")
        log.info(all_filenames)

        log.info("********** create hospital")
        counter = Counter()
        counter.set_with_database(database=self.database)

        all_metadata = read_tabular_file_as_string(self.execution.metadata_filepath)  # keep all metadata as str
        first = True
        for one_filename in all_filenames:
            if ".vcf" in one_filename:
                # we probably have a regex for specifying the VCF location (*.vcf or my_folder/*.vcf)
                # we will process them during the pre-processing of the data
                pass
            elif one_filename != "":
                log.info(one_filename)

                # set the current filepath
                self.execution.current_filepath = os.path.join(DOCKER_FOLDER_DATA, one_filename)
                log.info(self.execution.current_filepath)

                # create a new Dataset instance
                counter.set_with_database(database=self.database)
                dataset = Dataset(identifier=NO_ID, database=self.database, docker_path=self.execution.current_filepath, version_notes=None, license=None, counter=counter)
                self.datasets.append(dataset)
                self.execution.current_dataset_gid = dataset.global_identifier

                # create the hospital only once
                if first:
                    self.create_hospital(counter=counter, dataset_id=dataset.identifier)
                    first = False

                # get metadata of file
                log.info(one_filename)
                if one_filename not in all_metadata[MetadataColumns.DATASET_NAME].unique():
                    raise ValueError(f"The current dataset ({one_filename}) is not described in the provided metadata file.")
                else:
                    log.info(f"--- Extract metadata for file '{self.execution.current_filepath}', with number {self.execution.current_file_number}")
                    metadata = pd.DataFrame(all_metadata[all_metadata[MetadataColumns.DATASET_NAME].values == one_filename])

                    log.info(f"--- Starting to transform file '{self.execution.current_filepath}', with number {self.execution.current_file_number}")
                    # we have to iterate over all profiles associated to the dataset because we cannot do it in the Extract
                    # because the transform and load have to be applied on each pair (ds, profile)
                    # the Extract class will take care of setting the profile if it is associated to the current dataset
                    unique_profiles_of_current_dataset = pd.unique(metadata[MetadataColumns.PROFILE])
                    count_profiles = 0
                    for profile in unique_profiles_of_current_dataset:
                        count_profiles += 1
                        profile = Profile.normalize(profile)
                        log.info(f"using profile {profile}")

                        # check whether this is the last profile of the last dataset
                        # to know whether we should compute indexes
                        if dataset.identifier == len(all_filenames) and count_profiles == len(unique_profiles_of_current_dataset):
                            compute_indexes = True

                        # EXTRACT
                        time_stats.start(dataset=dataset.global_identifier, key=TimerKeys.EXTRACT_TIME)
                        self.extract = Extract(metadata=metadata, profile=profile, database=self.database, execution=self.execution, quality_stats=quality_stats)
                        self.extract.run()
                        time_stats.increment(dataset=dataset.global_identifier, key=TimerKeys.EXTRACT_TIME)

                        if self.extract.metadata is not None:
                            log.info(f"running transform on dataset {self.execution.current_filepath} with profile {profile}")
                            # TRANSFORM
                            time_stats.start(dataset=dataset.global_identifier, key=TimerKeys.TRANSFORM_TIME)
                            self.transform = Transform(database=self.database, execution=self.execution, data=self.extract.data,
                                                       metadata=self.extract.metadata,
                                                       mapping_column_to_categorical_value=self.extract.mapping_column_to_categorical_value,
                                                       mapping_column_to_unit=self.extract.mapping_column_to_unit,
                                                       mapping_column_to_domain=self.extract.mapping_column_to_domain,
                                                       mapping_column_to_type=None,  # this will be computed during the Transform step
                                                       profile=profile, load_patients=count_profiles == 1,
                                                       dataset_id=dataset.identifier, dataset_key=dataset,
                                                       quality_stats=quality_stats)
                            self.transform.run()
                            time_stats.increment(dataset=dataset.global_identifier, key=TimerKeys.TRANSFORM_TIME)

                            # LOAD
                            time_stats.start(dataset=dataset.global_identifier, key=TimerKeys.LOAD_TIME)
                            # log.info(f"{one_filename} -> {compute_indexes}")
                            # create indexes only if this is the last file (otherwise, we would create useless intermediate indexes)
                            self.load = Load(database=self.database, execution=self.execution, create_indexes=compute_indexes,
                                             dataset_id=dataset.identifier, profile=profile,
                                             quality_stats=quality_stats)
                            self.load.run()
                            time_stats.increment(dataset=dataset.global_identifier, key=TimerKeys.LOAD_TIME)
                self.execution.current_file_number += 1

        # save the datasets in the DB
        log.info(len(self.datasets))
        log.info(self.datasets[0])
        log.info(self.datasets)
        if self.database is not None and len(self.datasets) > 0:
            log.info([dataset.to_json() for dataset in self.datasets])
            self.database.upsert_one_batch_of_tuples(table_name=TableNames.DATASET, unique_variables=["docker_path"], the_batch=[dataset.to_json() for dataset in self.datasets], ordered=False)
            # for Dataset entity only, we create an index on the global identifier
            self.database.create_unique_index(table_name=TableNames.DATASET, columns={"global_identifier": 1})
        # compute their profiles
        log.info("profile computation")
        self.profile_computation = FeatureProfileComputation(database=self.database)
        self.profile_computation.compute_features_profiles()
        # compute DB stats
        db_stats = DatabaseStatistics(record_stats=True)
        db_stats.compute_stats(database=self.database)
        # compute the final report with all the stats
        self.reporting = Reporting(database=self.database, execution=self.execution, quality_stats=quality_stats, time_stats=time_stats, db_stats=db_stats)
        self.reporting.run()

    def create_hospital(self, counter: Counter, dataset_id: int) -> None:
        log.info(f"create hospital instance in memory")
        cursor = self.database.find_operation(table_name=TableNames.HOSPITAL, filter_dict={Hospital.NAME_: self.execution.hospital_name}, projection={})
        hospital_exists = False
        for _ in cursor:
            # the hospital already exists within the database, we do nothing
            # the ETL will take care of retrieving the existing hospital ID while creating records
            hospital_exists = True
        if not hospital_exists:
            # the hospital does not exist because we have reset the database, we create a new one
            log.info(self.execution.hospital_name)
            new_hospital = Hospital(identifier=NO_ID, name=self.execution.hospital_name, counter=counter)
            log.info(new_hospital)
            hospitals = [new_hospital.to_json()]
            write_in_file(resource_list=hospitals, current_working_dir=self.execution.working_dir_current,
                          table_name=TableNames.HOSPITAL, is_feature=False, dataset_id=dataset_id, to_json=False)
            self.database.load_json_in_table(table_name=TableNames.HOSPITAL, unique_variables=[Hospital.NAME_], dataset_id=dataset_id)



import dataclasses

from entities.OntologyResource import OntologyResource
from entities.Resource import Resource
from enums.Visibility import Visibility


@dataclasses.dataclass()
class Feature(Resource):
    name: str
    ontology_resource: OntologyResource
    data_type: str
    unit: str
    description: str
    categories: list[OntologyResource]
    visibility: Visibility
    dataset: str
    domain: dict

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    NAME_ = "name"
    ONTO_ = "ontology_resource"
    DT_ = "data_type"
    UNIT_ = "unit"
    DESCR_ = "description"
    CATEGORIES_ = "categories"
    VISIBILITY_ = "visibility"
    DOMAIN_ = "domain"

    # keys to be used when using an entity attribute as a query variable
    NAME__ = f"${NAME_}"
    ONTO__ = f"${ONTO_}"
    DT__ = f"${DT_}"
    UNIT__ = f"${UNIT_}"
    DESCR__ = f"${DESCR_}"
    CATEGORIES__ = f"${CATEGORIES_}"
    VISIBILITY__ = f"${VISIBILITY_}"
    DOMAIN__ = f"${DOMAIN_}"

    def __post_init__(self):
        super().__post_init__()

        # set up the feature attributes
        self.datasets = [self.dataset]
        if self.categories is not None and len(self.categories) == 0:
            self.categories = None  # this avoids to store empty arrays when there is no categorical values for a certain Feature
        if self.domain is not None and len(self.domain) == 0:
            self.domain = None



import os.path
from functools import reduce

import pandas as pd
from pandas import DataFrame

from database.Execution import Execution
from enums.DataTypes import DataTypes
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.Profile import Profile
from enums.Visibility import Visibility
from preprocessing.Preprocess import Preprocess
from utils.setup_logger import log


class PreprocessImgge(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str, all_columns_dataset: list):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)
        self.metadata = metadata
        self.all_columns_dataset = all_columns_dataset

    def preprocess(self):
        log.info("pre-process IMGGE data")

        if self.profile == Profile.PHENOTYPIC:
            hpo_data_column = MetadataColumns.normalize_name(column_name="hpo_data")

            # 0. replace the "|" by "," in the list of HPOs + normalize HPO labels
            self.data[hpo_data_column] = self.data[hpo_data_column].apply(lambda x: MetadataColumns.normalize_name(column_name=x.replace("|", ",").replace("_", ":")).replace("hp:", "").replace(" ", ""))

            # 1. flatten the list of HPO terms to obtain one column for each term. The ontology resource associated to each is an hpo term
            # see issue 305: https://git.rwth-aachen.de/padme-development/external/better/data-cataloging/etl/-/issues/305
            all_hpo_terms = []
            for one_hpo_list in self.data[hpo_data_column]:
                all_hpo_terms.extend(one_hpo_list.split(","))
            log.info(f"{len(all_hpo_terms)} HPO terms")
            all_hpo_terms = list(set(all_hpo_terms))
            all_hpo_terms = [MetadataColumns.normalize_name(column_name=hpo_term) for hpo_term in all_hpo_terms]
            log.info(f"{len(all_hpo_terms)} distinct HPO terms")

            # build the set of columns to add and add it once to avoid fragmented dataframe/performance issues
            # we need to write "0" and "1" instead of True and False because the df is read as a string
            hpo_dfs = [pd.Series(self.data[hpo_data_column].apply(lambda x: "1" if hpo_term in x else "0"), name=hpo_term).to_frame() for hpo_term in all_hpo_terms]
            indexed_hpo_dfs = []
            for one_hpo_df in hpo_dfs:
                one_hpo_df[self.execution.patient_id_column_name] = self.data[self.execution.patient_id_column_name]
                indexed_hpo_dfs.append(one_hpo_df)
            concat_hpo_dfs = reduce(lambda x, y: x.merge(y, how="inner", on=self.execution.patient_id_column_name), indexed_hpo_dfs)  # concat_hpo_dfs = pd.merge(indexed_hpo_dfs, axis=1)
            self.data = self.data.merge(concat_hpo_dfs, how="inner", on=self.execution.patient_id_column_name)

            # 2. then, add the flattened HPO terms as features in the metadata
            current_filename = os.path.basename(self.execution.current_filepath)
            hpo_metadata = DataFrame(data={
                MetadataColumns.ONTO_NAME: [Ontologies.HPO["name"] for _ in range(len(hpo_dfs))],
                MetadataColumns.ONTO_CODE: all_hpo_terms,
                MetadataColumns.DATASET_NAME: [current_filename for _ in range(len(hpo_dfs))],
                MetadataColumns.COLUMN_NAME: all_hpo_terms,
                MetadataColumns.PROFILE: [Profile.PHENOTYPIC for _ in range(len(hpo_dfs))],
                MetadataColumns.VISIBILITY: [Visibility.PUBLIC for _ in range(len(hpo_dfs))],
                MetadataColumns.ETL_TYPE: [DataTypes.BOOLEAN for _ in range(len(hpo_dfs))]
            })

            self.metadata = pd.concat([self.metadata, hpo_metadata], axis=0)
            self.all_columns_dataset = pd.concat([self.all_columns_dataset, pd.Series(all_hpo_terms)])



import re
from datetime import datetime

import numpy as np
import requests


API_SESSION = requests.Session()

# 1000 is the max supported by Mongodb.
# If we set BATCH_SIZE > 1000, Mongodb will send batch of 1k instances,
# but we will still have fewer files to write and read, thus increasing performance
BATCH_SIZE = 10000
MAX_FILE_SIZE = 15 * 1048576  # 10Mb = 10x 1024^2 (bytes) -- MongoDB is limited to 16Mo but our size estimate is usually lower than the real size so we stay at 10Mo to avoid creating files with a real size larger than 16Mo

PATTERN_VALUE_UNIT = re.compile(r'^ *([0-9]+[.,]*[0-9]*) *([a-zA-Z_.-]+) *$')  # we add start and end delimiters (^ and $) to not process cells with multiples values inside

SNOMED_OPERATORS_LIST = ["|", "(", ")", "{", "}", ",", ":", "=", "+"]
SNOMED_OPERATORS_STR = "".join(SNOMED_OPERATORS_LIST)

DEFAULT_ONTOLOGY_RESOURCE_LABEL = ""

DATASET_GLOBAL_IDENTIFIER_PREFIX = "http://better-health-project.eu/datasets/"

NO_ID = -1

NAN_VALUES = ["no information", "-", "nan", "na", "none", "n/a", "null", "np.nan", np.nan, "/"]
DEFAULT_NAN_VALUE = np.nan

PRINT_QUERIES = True



from database.Counter import Counter


class TestCounter:
    def test_constructor(self):
        counter = Counter()
        assert counter.resource_id == 0

    def test_incr_simple(self):
        counter = Counter()
        counter.increment()
        assert counter.resource_id == 1

    def test_incr_multiple(self):
        counter = Counter()
        for _ in range(10):
            counter.increment()
        assert counter.resource_id == 10

    def test_set(self):
        counter = Counter()
        counter.set(new_value=100)
        assert counter.resource_id == 100

        counter.increment()
        assert counter.resource_id == 101

    def test_reset(self):
        counter = Counter()
        counter.increment()
        counter.increment()
        assert counter.resource_id == 2
        counter.reset()
        assert counter.resource_id == 0



import dataclasses
import json
from datetime import datetime

from constants.methods import factory


@dataclasses.dataclass()
class MainStatistics:
    record_stats: bool
    timestamp: datetime = dataclasses.field(init=False)

    def __post_init__(self):
        self.timestamp = datetime.now()

    def to_json(self):
        return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



import os

import pandas as pd
from pandas import DataFrame

from database.Execution import Execution
from enums.DataTypes import DataTypes
from enums.MetadataColumns import MetadataColumns
from enums.Ontologies import Ontologies
from enums.Profile import Profile
from enums.Visibility import Visibility
from preprocessing.Preprocess import Preprocess
from utils.setup_logger import log


class PreprocessLafe(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str, all_columns_dataset: list):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)

        self.metadata = metadata
        self.all_columns_dataset = all_columns_dataset

    def preprocess(self):
        log.info("pre-process LAFE data")
        log.info(self.data)
        log.info(self.data.columns)
        normalized_columns = [MetadataColumns.normalize_name(column_name=column) for column in self.data.columns]
        log.info(self.execution.current_filepath)
        log.info(self.profile)
        log.info(self.execution.sample_id_column_name)
        if self.profile == Profile.CLINICAL and "Dynamic_Clinical_Table" in self.execution.current_filepath and self.execution.sample_id_column_name not in normalized_columns:
            # 1. add sample IDs for the data file "Dynamic_Clinical_Table.xlsx"
            self.execution.sample_id_column_name = "sample_id"
            self.data[self.execution.sample_id_column_name] = [i for i in range(1, len(self.data)+1)]
            log.info(self.data.columns)
            log.info(self.data)

            # 2. update the metadata to make the sample ID column appear
            df_line_sid = DataFrame({
                MetadataColumns.ONTO_NAME: [Ontologies.LOINC["name"]],
                MetadataColumns.ONTO_CODE: ["57723-9"],
                MetadataColumns.COLUMN_NAME: [self.execution.sample_id_column_name],
                MetadataColumns.PROFILE: [self.profile],
                MetadataColumns.DATASET_NAME: [os.path.basename(self.execution.current_filepath)],
                MetadataColumns.ETL_TYPE: [DataTypes.STRING],
                MetadataColumns.VISIBILITY: [Visibility.PUBLIC]
            })
            log.info(df_line_sid)
            log.info(self.metadata)
            self.metadata = pd.concat([self.metadata, df_line_sid])
            log.info(self.metadata)
            log.info(self.all_columns_dataset)
            self.all_columns_dataset = pd.concat([self.all_columns_dataset, pd.Series(data=[self.execution.sample_id_column_name])])
            log.info(self.all_columns_dataset)



import dataclasses
import json

import pandas as pd
from pandas import DataFrame
from pymongo import MongoClient

from database.Operators import Operators
from entities.OntologyResource import OntologyResource
from enums.Ontologies import Ontologies
from enums.QueryTypes import QueryTypes
from enums.TableNames import TableNames
from utils.setup_logger import log


@dataclasses.dataclass()
class DataRetriever:
    # db connection
    mongodb_url: str  # "mongodb://localhost:27018/"
    db_name: str
    query_type: QueryTypes  # METADATA or DATA

    # user input to build the query to retrieve feature data (metadata)
    feature_list: list = dataclasses.field(default_factory=list)  # user input of the form ["8116006:278201002=(\"HPO\",288467006)", "12953007", ...]

    # user input to build the query to retrieve record data
    # FEATURE_CODES = {"hypotonia": "8116006:278201002=(\"HPO\",288467006)", "vcf_path": "12953007"}
    feature_selected: dict = dataclasses.field(default_factory=list)  # user input of the form {"user variable name": "ontology code"
    # FEATURES_VALUE_PROCESS = {"hypotonia": {"$addFields": { "hypotonia": { "$in": ["hp:0001252", "$hypotonia"] }}}, "vcf_path": None}
    feature_value_process: dict = dataclasses.field(default_factory=list)  # user input of the form {"user variable name": { the mongodb operation or None } }
    feature_filters: dict = dataclasses.field(default_factory=list)  # user input of the form {"user variable name": {"code": "ontology code", "filter": { the mongodb operator }}}
    features_info: dict = dataclasses.field(init=False)  # generated from the 3 user input (select, preprocess, and filters)
    the_dataframe: DataFrame = dataclasses.field(init=False)

    def __post_init__(self):
        self.client = MongoClient(host=self.mongodb_url, serverSelectionTimeoutMS=30, w=0, directConnection=True)
        log.info(self.client)
        self.db = self.client[self.db_name]
        log.info(self.db)
        self.the_query = ""

        # check that the query type match the given parameters
        if self.query_type == QueryTypes.METADATA:
            if self.feature_list is None or len(self.feature_list) == 0:
                raise Exception("You specified that you want to retrieve metadata from the database but did not specify which features to retrieve.")
            else:
                new_feature_list = []
                for code in self.feature_list:
                    code_split = code.split(":", 1)  # split on the first :
                    system, code = Ontologies.get_enum_from_name(Ontologies.normalize_name(code_split[0])), Ontologies.normalize_code(code_split[1])
                    new_feature_list.append(OntologyResource(system=system, code=code, label=" ", quality_stats=None))
                self.feature_list = new_feature_list
        elif self.query_type == QueryTypes.DATA:
            if self.feature_selected is None or len(self.feature_selected) == 0:
                raise Exception("You specified that you want to retrieve data from the database but did not specify which features to retrieve.")
            else:
                # normalize ontology resource codes
                # we have to set the system to no ontology, otherwise no ontology resource will be created
                # we have to set the label to the non-empty string, otherwise a label wil be computed
                # OntologyResource(system=Ontologies.NONE, code=code, label=" ", quality_stats=None).code
                self.features_info = {}
                i = 1
                for key, code in self.feature_selected.items():
                    code_split = code.split(":", 1)  # split on the first :
                    system, code = Ontologies.get_enum_from_name(Ontologies.normalize_name(code_split[0])), Ontologies.normalize_code(code_split[1])
                    self.features_info[key] = {
                        "the_or": OntologyResource(system=system, code=code, label=" ", quality_stats=None),
                        "filter": None,
                        "process": None,
                        "show": True,
                        "i": i
                    }
                    i = i + 1
                log.info(self.features_info)
                # for each feature specified in the filter, add it to the infos
                for feature_name, feature in self.feature_filters.items():
                    if feature_name not in self.features_info:
                        self.features_info[feature_name] = {}
                        code_split = self.feature_filters[feature_name]["code"].split(":", 1)  # split on the first :
                        system, code = Ontologies.get_enum_from_name(Ontologies.normalize_name(code_split[0])), Ontologies.normalize_code(code_split[1])
                        self.features_info[feature_name]["the_or"] = OntologyResource(system=system, code=code, label=" ", quality_stats=None)
                        self.features_info[feature_name]["i"] = i
                        i = i + 1
                        self.features_info[feature_name]["show"] = False
                    if feature_name not in self.feature_filters:
                        pass
                    else:
                        if self.feature_filters[feature_name] is not None and "filter" in self.feature_filters[feature_name]:
                            self.features_info[feature_name]["filter"] = self.feature_filters[feature_name]["filter"]
                        else:
                            self.features_info[feature_name]["filter"] = None
                    self.features_info[feature_name]["process"] = None
                # for each feature specified in the post-process, add it to the info
                for feature_name, process in self.feature_value_process.items():
                    if feature_name not in self.features_info:
                        # we do not have a code for this feature
                        pass
                    else:
                        self.features_info[feature_name]["process"] = process
                log.info(self.features_info)

    def run(self):
        if self.query_type == QueryTypes.DATA:
            self.retrieve_records()
        elif self.query_type == QueryTypes.METADATA:
            self.retrieve_features()

    def retrieve_features(self):
        log.info("***************")
        log.info("Creating indexes")
        self.db[TableNames.FEATURE].create_index("ontology_resource.system")
        self.db[TableNames.FEATURE].create_index("ontology_resource.code")

        log.info("Generating the query")
        self.generate_metadata_query()
        log.info(self.the_query)
        log.info("Creating the dataframe")
        self.the_dataframe = pd.DataFrame(self.db[TableNames.FEATURE].aggregate(json.loads(self.the_query))).set_index("identifier")
        log.info("Done.")

    def generate_metadata_query(self):
        # {"$or": [{"$and": [{"system": "a"}, {"code": 1}]}, {"$and": [{"system": "b", "code": 2}]}]}
        match_feature_codes = Operators.or_operator(list_of_conditions=[
            Operators.and_operator(list_of_conditions=[{"ontology_resource.system": ontology_resource.system, "ontology_resource.code": ontology_resource.code}]) for ontology_resource in self.feature_list
        ])

        self.the_query += "["
        self.the_query += json.dumps(Operators.match(field=None, value=match_feature_codes, is_regex=False))
        self.the_query += ","
        self.the_query += json.dumps(Operators.set_variables([{"name": "onto_system", "operation": "$ontology_resource.system"}, {"name": "onto_code", "operation": "$ontology_resource.code"}]))
        self.the_query += ","
        self.the_query += json.dumps(Operators.unset_variables(["_id"]))
        self.the_query += "]"

    def retrieve_records(self):
        feature_codes_inv = {v["the_or"].to_string(): k for k, v in self.features_info.items()}
        i = 1
        # get features for features asked in the SELECT
        match_feature_codes = Operators.or_operator(list_of_conditions=[
            Operators.and_operator(list_of_conditions=[{"ontology_resource.system": feature["the_or"].system,
                                                        "ontology_resource.code": feature["the_or"].code}]) for
            feature_name, feature in self.features_info.items() #if feature["show"] is True
        ])
        log.info(match_feature_codes)

        for res in self.db[TableNames.FEATURE].find(match_feature_codes):
            log.info(res)
            log.info(res['ontology_resource']['system'])
            log.info(res['ontology_resource']['code'])
            onto_system = res["ontology_resource"]["system"]
            onto_code = res["ontology_resource"]["code"]
            total = str(onto_system) + ":" + str(onto_code)
            feature_user_name = feature_codes_inv[total]
            self.features_info[feature_user_name]["identifier"] = res["identifier"]
            i = i + 1
        log.info(self.features_info)

        log.info("***************")
        log.info("Creating indexes")
        self.db[TableNames.RECORD].create_index("instantiates")
        self.db[TableNames.FEATURE].create_index("ontology_resource.code")

        log.info("Generating the query")
        self.generate_data_query()
        log.info(self.the_query)
        log.info("Creating the dataframe")
        self.the_dataframe = pd.DataFrame(self.db[TableNames.RECORD].aggregate(json.loads(self.the_query))).set_index(
            "has_subject")
        log.info("Done.")

    def generate_data_query(self):
        # 0. compute the lookups
        self.rec_internal(position=1)
        # we now need to add the latest stages to set final variables from the lookups
        # this cannot be done during the recursion, so we do it now
        self.the_query = self.the_query[0:len(self.the_query) - 1]  # remove the last ] before adding final stages
        # log.info(self.the_query)
        self.the_query += ","

        # 1. set variables of the unwind to place them on the first level of the computed record
        set_variables = []
        for feature_name, feature in self.features_info.items():
            lookup_names = "$"
            for j in range(1, feature["i"]):
                lookup_names += f"lookup_{j}."
            lookup_names += "value"
            set_variables.append({"name": feature_name, "operation": lookup_names})
        self.the_query += json.dumps(Operators.set_variables(variables=set_variables))
        self.the_query += ","

        # 2. filter the obtained records
        for feature in self.features_info.values():
            if "filter" in feature and feature["filter"] is not None:
                for select_key, select_operator in feature["filter"].items():
                    self.the_query += json.dumps(Operators.match(field=select_key, value=select_operator, is_regex=False)) + ","

        # 3. set the value to keep + those to flatten
        for feature_name, feature in self.features_info.items():
            if "process" in feature and feature["process"] is not None:
                if feature["process"] == "get_label":
                    # we are in a category, we get its label
                    self.the_query += json.dumps(Operators.set_variables(variables=[{"name": feature_name, "operation": f"${feature_name}.label"}]))
                    self.the_query += ","
                else:
                    # the user gave a personalized post-process, we apply it as is
                    self.the_query += json.dumps(feature["process"])
                    self.the_query += ","
            log.info(self.the_query)

        # 4. project only variables specified in the SELECT (not those only in the WHERE)
        projected_values = {feature_name: 1 for feature_name, feature in self.features_info.items() if feature["show"] is True}
        projected_values["_id"] = 0
        projected_values["has_subject"] = 1
        self.the_query += json.dumps(Operators.project(field=None, projected_value=projected_values))
        self.the_query += "]"
        log.info(self.the_query)

    def rec_internal(self, position):
        log.info(f"rec {position} with {self.features_info}")
        feature_id = [feature["identifier"] for feature_name, feature in self.features_info.items() if feature["i"] == position][0]
        if position == len(self.features_info):
            self.the_query += "["
            # log.info(self.the_query)
            self.the_query += json.dumps(Operators.match(field="instantiates", value=feature_id, is_regex=False))
            # log.info(self.the_query)
            self.the_query += ","
            # log.info(self.the_query)
            # if position > 1 and position < max_position:
            self.the_query += json.dumps(
                Operators.match(field=None, value={"$expr": {"$eq": [f"$$id_{position - 1}", "$has_subject"]}},
                                is_regex=False))
            # log.info(self.the_query)
            self.the_query += "]"
            # log.info(self.the_query)
        else:
            self.the_query += "["
            # log.info(self.the_query)
            self.the_query += json.dumps(Operators.match(field="instantiates", value=feature_id, is_regex=False))
            # log.info(self.the_query)
            self.the_query += ","
            # log.info(self.the_query)
            if position > 1:
                self.the_query += json.dumps(
                    Operators.match(field=None, value={"$expr": {"$eq": [f"$$id_{position - 1}", "$has_subject"]}},
                                    is_regex=False))
                # log.info(self.the_query)
                self.the_query += ","
                # log.info(self.the_query)

            # the lookup needs to be only partially written in order to compute its internal pipeline
            # this is why we need to write manually the string in order to compute the internal pipeline in-between
            self.the_query += "{\"$lookup\": {\"from\": \"" + TableNames.RECORD + "\", \"let\": {\"id_" + str(
                position) + "\": \"$has_subject\"}, \"as\": \"lookup_" + str(position) + "\", \"pipeline\": "
            # log.info(self.the_query)
            self.rec_internal(position + 1)
            # log.info(self.the_query)
            self.the_query += "}}"
            # log.info(self.the_query)
            self.the_query += ","
            # log.info(self.the_query)
            self.the_query += json.dumps(Operators.unwind(field=f"lookup_{position}"))
            # log.info(self.the_query)
            self.the_query += "]"
            # log.info(self.the_query)



from __future__ import annotations

import dataclasses
import re
from urllib.parse import quote

from constants.defaults import SNOMED_OPERATORS_LIST, DEFAULT_ONTOLOGY_RESOURCE_LABEL, SNOMED_OPERATORS_STR
from constants.methods import factory
from enums.AccessTypes import AccessTypes
from enums.Ontologies import Ontologies
from main_statistics.QualityStatistics import QualityStatistics
from utils.api_utils import send_query_to_api, parse_xml_response, parse_json_response, parse_html_response
from utils.setup_logger import log
from utils.str_utils import remove_specific_tokens, process_spaces, remove_operators_in_strings


@dataclasses.dataclass()
class OntologyResource:
    system: dict
    code: str
    label: str | None
    quality_stats: QualityStatistics | None = dataclasses.field(repr=False)
    show_warning: bool = dataclasses.field(repr=False, default=True)

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    SYSTEM_ = "system"
    CODE_ = "code"
    LABEL_ = "label"

    def __post_init__(self):
        # every attribute that is there will be serialized in the Ontology Resource
        # to avoid this, one needs to explicitly say which attributes are to be removed from the JSON serialization
        # using the __get_state__ method
        self.quality_stats = self.quality_stats if self.quality_stats is not None else QualityStatistics(record_stats=False)
        if len(self.system) == 0 or len(self.code) == 0:
            # no ontology code has been provided for that variable name, let's skip it
            # the only case when we don't want to skip it is when a label is provided as input
            if self.label is not None and self.label != "":
                if self.show_warning:
                    # self.show_warning -> False if we are building OR from the database, True otherwise (we are building OR in the code)
                    log.warning(f"Creating an OntologyResource with a label only (label={self.label}).")
            else:
                if self.show_warning:
                    # self.show_warning -> False if we are building OR from the database, True otherwise (we are building OR in the code)
                    log.warning("Could not create an OntologyResource with no ontology system and/or code.")
        else:
            # this corresponds to the first (and only) ontology system;
            # if there are many, we record only the first but make API calls with all
            # this is because an Ontology can have a single system
            # TODO Nelly: maybe there is a better way?
            self.system = self.system["url"]
            self.code = self.code.replace("ORPHA:", "").replace("orpha:", "").replace("GO:", "").replace("go:", "")
            self.code = process_spaces(input_string=self.code)
            self.code = re.sub(r" *([" + SNOMED_OPERATORS_STR + "]+) *", r"\1", self.code)  # remove spaces around operators; r"\1" means: replace with first captured group
            self.code = remove_operators_in_strings(input_string=self.code)  # for every label inside |, '' or "", we remove possible operators
            code_elements = self.compute_elements(full_code=self.code)
            self.compute_code(code_elements=code_elements)
            if self.label is None:
                # when we create a new OntologyResource from scratch, we need to compute the label with ontology API
                # if the query to the API does not work, we can still use the column name as the label of the OntoResource
                self.compute_label(code_elements=code_elements, quality_stats=self.quality_stats)

    def compute_elements(self, full_code: str) -> list:
        elements = []
        regex_elements = re.split(r"(?=["+SNOMED_OPERATORS_STR+"])|(?<=["+SNOMED_OPERATORS_STR+"])", full_code)
        # now self.elements may still contain spaces
        # therefore, we remove them manually afterward
        for element in regex_elements:
            if element == "" or element is None or element == " ":
                # the regex sometimes returns empty or None elements, we skip them
                # it may also identify spaces around operators, we skip them too
                pass
            else:
                elements.append(element)
        return elements

    def compute_code(self, code_elements: list) -> None:
        self.code = ""
        for i in range(len(code_elements)):
            element = code_elements[i]
            if element not in SNOMED_OPERATORS_LIST:
                # this is not an operator
                if i-1 >= 0 and code_elements[i-1] == "|":
                    # we are in an annotation, thus we skip it
                    pass
                elif element.startswith("\""):
                    # this is a constant, thus we keep it and only process space (not caps)
                    self.code += process_spaces(input_string=element)
                else:
                    # this is a code
                    self.code += Ontologies.normalize_code(ontology_code=element)
            else:
                # this is an operator, we add it only if this in not the pipe
                if element != "|":
                    self.code += element

    def compute_label(self, code_elements: list, quality_stats: QualityStatistics) -> None:
        self.label = ""
        for i in range(len(code_elements)):
            element = code_elements[i]
            if element not in SNOMED_OPERATORS_LIST and not element.startswith("\""):  # " is used for surrounding constants, e.g., "HPO" in 278201002|Classification|="HPO"
                # this is not an operator, nor a constant
                if i - 1 >= 0 and code_elements[i - 1] == "|":
                    # we are in an annotation, thus we skip it
                    pass
                else:
                    # this is a code, we get its label (name)
                    resource_label = OntologyResource.get_resource_label_from_api(system=self.system, single_code=element, quality_stats=quality_stats)
                    if resource_label is not None:
                        resource_label = process_spaces(input_string=resource_label)
                        resource_label = remove_specific_tokens(input_string=resource_label, tokens=["(property)", "- finding", "-finding", "(qualifier value)", "(observable entity)", "(social concept)", "(procedure)", "(assessment scale)", "- action", "-action", "- attribute", "-attribute"])  # useless and may break parsing (due to parenthesis and dash)
                        resource_label = remove_specific_tokens(input_string=resource_label, tokens=SNOMED_OPERATORS_LIST)  # if we don't remove them, this will break future parsing
                        resource_label = process_spaces(input_string=resource_label)
                        self.label += resource_label
            else:
                # this is an operator, we add it only if this in not the pipe
                if element != "|":
                    self.label += element

    @classmethod
    def get_resource_label_from_api(cls, system: str, single_code: str, quality_stats: QualityStatistics) -> str:
        # column_name is to be used when the label of the OntologyResource could not be computed with any of the APIs
        compute_from_api = True
        if compute_from_api:
            try:
                if system == Ontologies.SNOMEDCT["url"]:
                    url_resource = quote(f"http://purl.bioontology.org/ontology/SNOMEDCT/{single_code}", safe="")
                    url = f"http://data.bioontology.org/ontologies/SNOMEDCT/classes/{url_resource}"
                    response = send_query_to_api(url=url, secret="d6fb9c05-3309-4158-892f-65434a9133b9", access_type=AccessTypes.API_KEY_IN_URL)
                    if response is None:
                        error = f"Failed connection to SNOMED-CT API."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if "prefLabel" in data:
                            return data["prefLabel"]
                        else:
                            error = f"No label field for resource {single_code}."
                    elif response.status_code == 404 or response.status_code == 400:
                        error = f"Resource {single_code} not found."
                    else:
                        error = f"Failed connection to SNOMED-CT API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.LOINC["url"]:
                    url = f"https://loinc.regenstrief.org/searchapi/loincs?query={single_code}"
                    response = send_query_to_api(url=url, secret="nbarret d7=47@xiz$g=-Ns", access_type=AccessTypes.AUTHENTICATION)
                    data = parse_json_response(response)
                    if response is None:
                        error = f"Failed connection to LOINC API."
                    elif "Results" in data:
                        if len(data["Results"]) > 0:
                            if "COMPONENT" in data["Results"][0]:
                                return data["Results"][0]["COMPONENT"]
                            else:
                                error = f"No field label for resource {single_code}"
                        else:
                            error = f"Resource {single_code} not found."
                    else:
                        error = f"Failed connection to LOINC API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.PUBCHEM["url"]:
                    url = f"https://pubchem.ncbi.nlm.nih.gov/rest/pug/compound/cid/{single_code}/description/JSON"
                    response = send_query_to_api(url, secret=None, access_type=AccessTypes.USER_AGENT)
                    if response is None:
                        error = f"Failed connection to PUBCHEM API."
                    elif response.status_code == 404 or response.status_code == 400:
                        error = f"Resource {single_code} not found."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if "InformationList" in data and "Information" in data["InformationList"] and len(data["InformationList"]["Information"]) > 0:
                            if "Title" in data["InformationList"]["Information"][0]:
                                return data["InformationList"]["Information"][0]["Title"]
                            else:
                                error = f"No label field for resource {single_code}."
                        else:
                            error = f"Resource {single_code} not found."
                    else:
                        error = f"Failed connection to PUBCHEM API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.CLIR["url"]:
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error="No API access for the CLIR ontology.")
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.GSSO["url"]:
                    iri = f"http://purl.obolibrary.org/obo/{single_code.upper()}"  # we need to upper case the GSSO_, otherwise the API returns None
                    url = f"https://ontobee.org/ontology/GSSO?iri={iri}"
                    response = send_query_to_api(url=url, secret="d6fb9c05-3309-4158-892f-65434a9133b9", access_type=AccessTypes.API_KEY_IN_BEARER)
                    if response is None:
                        error = f"Failed connection to GSSO API."
                    elif response.status_code == 200:
                        data = parse_xml_response(response)  # data is an XML document
                        classes = data.getElementsByTagName('Class')
                        # <Class rdf:about="http://purl.obolibrary.org/obo/GSSO_006450">
                        #   <rdfs:label xml:lang="en">individual bullying</rdfs:label>
                        #   <rdfs:subClassOf rdf:resource="http://purl.obolibrary.org/obo/GSSO_000435" />
                        #   <ns2:IAO_0000115 xml:lang="en">Bullying perpetrated by a single person against a target ot targets.</ns2:IAO_0000115>
                        # </Class>
                        if len(classes) > 0:
                            expected_element = None
                            for one_class in classes:
                                if one_class.getAttribute("rdf:about") == iri:
                                    expected_element = one_class
                            if expected_element is None:
                                error = f"Resource {single_code} not found."
                            else:
                                if len(expected_element.getElementsByTagName("rdfs:label")) > 0:
                                    return expected_element.getElementsByTagName("rdfs:label")[0].childNodes[0].data
                                else:
                                    error = f"No field label for resource {single_code}"
                        else:
                            error = f"Resource {single_code} not found."
                    elif response.status_code == 404 or response.status_code == 400:
                        error = f"Resource {single_code} not found."
                    else:
                        error = f"Failed connection to GSSO API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.ORPHANET["url"]:
                    url = f"https://api.orphacode.org/EN/ClinicalEntity/orphacode/{single_code}/Name"
                    response = send_query_to_api(url=url, secret="nbarret", access_type=AccessTypes.API_KEY_IN_HEADER)
                    if response is None:
                        error = f"Failed connection to ORPHANET API."
                    elif response.status_code == 404 or response.status_code == 400:
                        error = f"Resource {single_code} not found."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if "Preferred term" in data:
                            return data["Preferred term"]
                        else:
                            error = f"No label field for resource {single_code}."
                    else:
                        error = f"Failed connection to ORPHANET API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.GENE_ONTOLOGY["url"]:
                    # as of 03/09/2024, this ontology is queried by accessing the webpage describing the resource
                    # it seems that there is an RDF query tool, but it is not sure that this can be queried as an API
                    # and there is no documentation on existing properties to query some codes
                    url = f"https://amigo.geneontology.org/amigo/term/GO:{single_code}"
                    response = send_query_to_api(url=url, secret=None, access_type=AccessTypes.USER_AGENT)
                    if response is None:
                        error = f"Failed connection to GO API."
                    elif response.status_code == 200:
                        data = parse_html_response(response)
                        header = data.select_one("div.page-header > h1").text
                        if header != "":
                            return header
                        else:
                            error = f"No label field for resource {single_code}."
                    elif response.status_code == 404 or response.status_code == 400:
                        error = f"Resource {single_code} not found."
                    else:
                        error = f"Failed connection to GO API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.OMIM["url"]:
                    # TODO NELLY: get OMIM API key (default one on OMIM website nfNEOscLNWWXdSmUoMLPPA is unauthorized)
                    url = f"https://api.omim.org/api/entry?mimNumber={single_code}&include=text&format=json"
                    response = send_query_to_api(url=url, secret="nfNEOscLNWWXdSmUoMLPPA", access_type=AccessTypes.API_KEY_IN_HEADER)
                    if response is None:
                        error = f"Failed connection to OMIM API."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if "text" in data:
                            return data["text"]
                        else:
                            error = f"No text field for resource {single_code}."
                    else:
                        error = f"Failed connection to OMIM API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.HGNC["url"]:
                    url = f"https://rest.ensembl.org/xrefs/id/{single_code}?external_db=HGNC;content-type=application/json;all_levels=1"
                    response = send_query_to_api(url=url, secret=None, access_type=AccessTypes.USER_AGENT)
                    if response is None:
                        error = f"Failed connection to HGNC API."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if len(data) > 0 and "description" in data[0]:
                            return data[0]["description"]
                        else:
                            error = f"No text field for resource {single_code}."
                    else:
                        error = f"Failed connection to HGNC API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                elif system == Ontologies.HPO["url"]:
                    url = f"https://clinicaltables.nlm.nih.gov/api/hpo/v3/search?terms=HP:{single_code}&sf=id,name"
                    response = send_query_to_api(url=url, secret=None, access_type=AccessTypes.USER_AGENT)
                    if response is None:
                        error = f"Failed connection to HPO API."
                    elif response.status_code == 200:
                        data = parse_json_response(response)
                        if len(data) >= 4:
                            for one_response in data[3]:
                                if one_response[0] == f"HP:{single_code}":  # we further check which term corresponds exactly to the code we asked
                                    return one_response[1]
                        else:
                            error = f"No text field for resource {single_code}."
                    else:
                        error = f"Failed connection to HPO API."
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=error)
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
                else:
                    quality_stats.add_failed_api_call(system=system, code=single_code, api_error=f"Unknown ontology {system}")
                    return DEFAULT_ONTOLOGY_RESOURCE_LABEL
            except Exception as e:
                # the API could not be queried, returning empty string
                quality_stats.add_failed_api_call(system=system, code=single_code, api_error=e.args[0])
                return DEFAULT_ONTOLOGY_RESOURCE_LABEL
        else:
            return DEFAULT_ONTOLOGY_RESOURCE_LABEL

    def to_json(self):
        return dataclasses.asdict(self, dict_factory=factory)

    def to_string(self):
        return f"{self.system}:{self.code}"

    @classmethod
    def from_json(cls, json_or: dict, quality_stats: QualityStatistics):  # returns an OntologyResource
        # fill a new OntologyResource instance with a JSON-encoded OntologyResource
        the_system = Ontologies.get_enum_from_url(json_or["system"]) if "system" in json_or else ""
        the_code = json_or["code"] if "code" in json_or else ""
        the_label = json_or["label"] if "label" in json_or else None
        return OntologyResource(system=the_system, code=the_code, label=the_label, quality_stats=quality_stats, show_warning=False)

    def __eq__(self, other):
        if not isinstance(other, OntologyResource):
            raise TypeError(f"Could not compare the current instance with an instance of type {type(other)}.")

        # we do not use the display  because this would lead to unequal instances
        # if provided descriptions differ from one hospital to another
        return self.system == other.system and self.code == other.code



from __future__ import annotations

import base64
import xml.dom.minidom

from bs4 import BeautifulSoup
from requests import Response

from constants.defaults import API_SESSION
from enums.AccessTypes import AccessTypes
from utils.setup_logger import log


def send_query(url: str, headers: dict | None) -> Response | None:
    try:
        if headers is None:
            return API_SESSION.get(url, verify=True)
        else:
            return API_SESSION.get(url, headers=headers, verify=True)
    except Exception:
        # if there is an SSL problem, trying the same query without the SSL certificate verification
        try:
            if headers is None:
                return API_SESSION.get(url, verify=False)
            else:
                return API_SESSION.get(url, headers=headers, verify=False)
        except Exception as e:
            log.info(e)
            return None


# API ACCESS
def send_query_to_api(url, secret: str | None, access_type: AccessTypes) -> Response | None:
    # secret may contain an api key or (joint) username and password
    headers = {
        "User-Agent": "python-requests/2.31.0",
        "Accept-Encoding": "gzip, deflate",
        "Accept": "*/*",
        "Connection": "keep-alive",
    }
    if access_type == AccessTypes.USER_AGENT:
        return send_query(url=url, headers=headers)

    elif access_type == AccessTypes.AUTHENTICATION:
        username = secret.split(" ")[0]
        password = secret.split(" ")[1]
        base64string = base64.b64encode(bytes('%s:%s' % (username, password), "ascii"))
        headers["Authorization"] = f'Basic {base64string.decode("utf-8")}'  # Make sure to prepend 'Bearer ' before your API key
        return send_query(url=url, headers=headers)

    elif access_type == AccessTypes.API_KEY_IN_HEADER:
        headers["apiKey"] = secret
        return send_query(url=url, headers=headers)

    elif access_type == AccessTypes.API_KEY_IN_BEARER:
        headers["Authorization"] = f"Bearer {secret}"
        return send_query(url=url, headers=headers)

    elif access_type == AccessTypes.API_KEY_IN_URL:
        url_with_apikey = f"{url}?apikey={secret}"
        return send_query(url=url_with_apikey, headers=None)
    else:
        # unknown access type
        return None


def parse_json_response(response):
    # we need to load x2 and to dump to have a "real" JSON dict, parseable by Python
    # otherwise, we have a JSON-like string or JSON-like text data
    if response is not None:
        return response.json()  # json.loads(json.dumps(json.loads(response.content)))
    else:
        return {}


def parse_xml_response(response):
    return xml.dom.minidom.parseString(response.content)


def parse_html_response(response):
    return BeautifulSoup(response.text, "html.parser")


def load_xml_file(filepath: str):
    return xml.dom.minidom.parse(filepath)



import dataclasses

from entities.Record import Record
from enums.Profile import Profile
from enums.TableNames import TableNames


@dataclasses.dataclass()
class ClinicalRecord(Record):
    base_id: str
    entity_type: str = f"{Profile.CLINICAL}{TableNames.RECORD}"



from data_retriever.DataRetriever import DataRetriever

if __name__ == '__main__':
    # ssh -L 27018:131.175.120.86:27018 barret@131.175.120.86
    # command to open SSH tunnel to access the databases on GENEX from my laptop
    # to be run before running this script

    # ssh -L 27018:131.175.120.88:27018 barret@131.175.120.88
    # command to open SSH tunnel to access the databases on GECO from my laptop
    # to be run before running this script

    # SJD
    FEATURE_CODES = {"label": "snomedct:398152000", "filename": "snomedct:12953007"}  # "face_hpo": '8116006:398302004=(278201002="HPO")', "gene": "82256003:734841007"}
    FEATURES_FILTERS = {"label": None, "filename": None}
    FEATURES_VALUE_PROCESS = {"label": None, "filename": None}
    dataRetriever = DataRetriever(mongodb_url="mongodb://131.175.120.88:27018/", db_name="better_sjd",
                                  query_type="data",
                                  feature_selected=FEATURE_CODES, feature_value_process=FEATURES_VALUE_PROCESS,
                                  feature_filters=FEATURES_FILTERS)
    dataRetriever.run()
    print(dataRetriever.the_dataframe)
    dataRetriever.the_dataframe.to_csv("table_sjd.csv")

    # # IMGGE
    # FEATURE_CODES = {"hypotonia": "hp:0001252", "vcf_path": "snomedct:12953007"}
    # FEATURES_FILTERS = {"sex": {"code": "snomedct:734000001", "filter": {"sex.label": "Female"}}}
    # FEATURES_VALUE_PROCESS = {"hypotonia": None, "vcf_path": None, "sex": "get_label"}
    # dataRetriever = DataRetriever(mongodb_url="mongodb://localhost:27018/", db_name="better_imgge",
    #                               query_type="data",
    #                               feature_selected=FEATURE_CODES, feature_value_process=FEATURES_VALUE_PROCESS,
    #                               feature_filters=FEATURES_FILTERS)
    # dataRetriever.run()
    # print(dataRetriever.the_dataframe)
    # dataRetriever.the_dataframe.to_csv("table_imgge.csv")
    #
    # # Metadata SJD
    # FEATURE_CODES = ["snomed ct:398152000", "snomedct:12953007", "snomedct:82256003:  734841007"]
    # dataRetriever = DataRetriever(mongodb_url="mongodb://localhost:27018/", db_name="better_hsjd",
    #                               query_type="metadata", feature_list=FEATURE_CODES)
    # dataRetriever.run()
    # print(dataRetriever.the_dataframe)
    # dataRetriever.the_dataframe.to_csv("metadata_sjd.csv")



from constants.defaults import NO_ID
from database.Counter import Counter
from entities.Feature import Feature
from entities.Hospital import Hospital
from entities.Resource import Resource
from enums.HospitalNames import HospitalNames
from enums.TableNames import TableNames


class TestHospital:

    def test_constructor(self):
        counter = Counter()
        hospital1 = Hospital(identifier=NO_ID, name=HospitalNames.TEST_H1, counter=counter)

        assert hospital1 is not None
        assert hospital1.identifier is not None
        assert hospital1.identifier == 1

    def test_to_json(self):
        counter = Counter()
        hospital1 = Hospital(identifier=NO_ID, name=HospitalNames.TEST_H1, counter=counter)
        hospital1_json = hospital1.to_json()

        assert hospital1_json is not None
        assert hospital1_json == {
            Resource.IDENTIFIER_: 1,
            Feature.NAME_: HospitalNames.TEST_H1,
            Resource.TIMESTAMP_: hospital1.timestamp,
            Resource.ENTITY_TYPE_: TableNames.HOSPITAL
        }



from constants.defaults import SNOMED_OPERATORS_LIST


def process_spaces(input_string: str) -> str:
    # this removes trailing spaces and replaces sequences of several spaces by only one
    return ' '.join(input_string.split())


def remove_specific_tokens(input_string: str, tokens: list) -> str:
    for token in tokens:
        input_string = input_string.replace(token, "")
    return input_string


def remove_operators_in_strings(input_string: str) -> str:
    pipe_double_quote_indices = [i if char in ["|", "\""] else -1 for i, char in enumerate(input_string)]
    pipe_double_quote_indices = [i for i in pipe_double_quote_indices if i != -1]

    i = 0
    if len(pipe_double_quote_indices) % 2 != 0 or len(pipe_double_quote_indices) == 0:
        # there is an operator (|, ', or ") which does not have its sibling
        # or there are no operators in that code, so we can return immediately
        return input_string
    else:
        final_text = input_string
        while i < len(pipe_double_quote_indices):
            first_op_index = pipe_double_quote_indices[i]
            second_op_index = pipe_double_quote_indices[i+1]
            text_between_op = input_string[(first_op_index+1): second_op_index]
            new_text_between_op = remove_specific_tokens(input_string=text_between_op, tokens=["(property)", "- finding", "-finding"])
            for operator in SNOMED_OPERATORS_LIST:
                new_text_between_op = new_text_between_op.replace(operator, " ")
            final_text = final_text.replace(text_between_op, new_text_between_op)
            i = i+2
        return final_text



from constants.defaults import NO_ID
from database.Counter import Counter
from entities.Patient import Patient
from enums.TableNames import TableNames
from utils.setup_logger import log


class TestPatient:
    def test_constructor(self):
        """
        Test whether the Patient constructor correctly assign IDs and the resource type.
        :return: None.
        """
        # this is a new Patient, thus with a new anonymised ID
        counter = Counter()
        patient1 = Patient(identifier=NO_ID, counter=counter)
        assert patient1.identifier is not None
        assert patient1.identifier == 1

        # this is an existing Patient, for which an anonymized ID already exists
        counter = Counter()
        patient1 = Patient(identifier=123, counter=counter)
        assert patient1.identifier is not None
        assert patient1.identifier == 123

    def test_to_json(self):
        counter = Counter()
        patient1 = Patient(identifier=NO_ID, counter=counter)
        patient1_json = patient1.to_json()

        log.info(patient1_json)

        assert patient1_json is not None
        assert "identifier" in patient1_json
        assert patient1_json["identifier"] == 1
        assert "timestamp" in patient1_json
        assert patient1_json["timestamp"] == patient1.timestamp
        assert "entity_type" in patient1_json
        assert patient1_json["entity_type"] == TableNames.PATIENT



from database.Database import Database
from database.Execution import Execution
from entities.Record import Record
from entities.Resource import Resource
from enums.DiagnosisColumns import DiagnosisColumns
from enums.Profile import Profile
from enums.TableNames import TableNames
from enums.TimerKeys import TimerKeys
from etl.Task import Task
from main_statistics.QualityStatistics import QualityStatistics
from main_statistics.TimeStatistics import TimeStatistics
from utils.setup_logger import log


class Load(Task):
    def __init__(self, database: Database, execution: Execution, create_indexes: bool,
                 dataset_id: int, profile: str, quality_stats: QualityStatistics):
        super().__init__(database=database, execution=execution, quality_stats=quality_stats)
        self.create_indexes = create_indexes
        self.dataset_id = dataset_id
        self.profile = profile

    def run(self) -> None:
        # Insert resources that have not been inserted yet, i.e., all Record instances
        log.debug(f"in the Load class")
        self.load_records()

        # if everything has been loaded, we can create indexes
        if self.create_indexes:
            self.create_db_indexes()

    def load_records(self) -> None:
        log.info(f"load {self.profile} records")
        # we need to have registered_by, has_subject and instantiates for sure
        # we also need entity_type because we cannot have two indexes, one for non-clinical (reg, subj, inst) and one for clinical (reg, subj, inst, bid)
        # we also need base_id for the same reason, the value will be null for non-clinical records and clinical records without sample information
        unique_variables = [Record.REG_BY_, Record.SUBJECT_, Record.INSTANTIATES_, Resource.ENTITY_TYPE_, Record.BASE_ID_]
        if self.profile == Profile.DIAGNOSIS:
            # we allow patients to have several diagnoses
            unique_variables.append(DiagnosisColumns.DISEASE_COUNTER)
        log.info(unique_variables)
        self.database.load_json_in_table(table_name=TableNames.RECORD, unique_variables=unique_variables, dataset_id=self.dataset_id)

    def create_db_indexes(self) -> None:
        log.info(f"Creating indexes.")

        count = 0

        # 1. for each resource type, we create an index on its "identifier" and its creation date "timestamp"
        for table_name in TableNames.data_tables():
            log.info(f"add index on id + timestamp + entity_type for table {table_name}")
            self.database.create_unique_index(table_name=table_name, columns={Resource.IDENTIFIER_: 1})
            self.database.create_non_unique_index(table_name=table_name, columns={Resource.TIMESTAMP_: 1})
            self.database.create_non_unique_index(table_name=table_name, columns={Resource.ENTITY_TYPE_: 1})
            count += 3

        # 2. next, we also create resource-wise indexes

        # for Feature instances, we create an index both on the ontology (system) and a code
        # this is because we usually ask for a code for a given ontology (what is a code without its ontology? nothing)
        self.database.create_non_unique_index(table_name=TableNames.FEATURE, columns={"ontology_resource.system": 1, "ontology_resource.code": 1})
        count += 1

        # for Record instances, we create an index per reference because we usually join each reference to a table
        self.database.create_non_unique_index(table_name=TableNames.RECORD, columns={Record.INSTANTIATES_: 1})
        self.database.create_non_unique_index(table_name=TableNames.RECORD, columns={Record.SUBJECT_: 1})
        self.database.create_non_unique_index(table_name=TableNames.RECORD, columns={Record.DATASET_: 1})
        count += 3
        # we cannot create an index on the base id because some records have it (the clinical ones)
        # while others do not have (imaging, phenotypic, etc.)
        # if has_base_id:
        #     self.database.create_non_unique_index(table_name=TableNames.RECORD, columns={"base_id": 1})
        #     count += 1

        log.info(f"Finished to create {count} indexes.")



import dataclasses
import json
from datetime import datetime

from constants.defaults import NO_ID
from constants.methods import factory
from database.Counter import Counter
from database.Operators import Operators


@dataclasses.dataclass()
class Resource:
    identifier: int
    counter: Counter
    timestamp: dict = dataclasses.field(init=False)

    # keys to be used when writing JSON or queries
    # those names have to exactly match the variables names declared in entity classes
    IDENTIFIER_ = "identifier"
    TIMESTAMP_ = "timestamp"
    ENTITY_TYPE_ = "entity_type"
    DATASET_ = "dataset"
    # keys to be used when using an entity attribute as a query variable
    IDENTIFIER__ = f"${IDENTIFIER_}"
    TIMESTAMP__ = f"${TIMESTAMP_}"
    ENTITY_TYPE__ = f"${ENTITY_TYPE_}"
    DATASET__ = f"${DATASET_}"
    # keys to be used for group by in queries
    DATASET___ = f"$_id.{DATASET_}"

    def __post_init__(self):
        if self.identifier == NO_ID:
            # we are creating a new instance, we assign it a new ID
            self.identifier = self.counter.increment()
        self.timestamp = Operators.from_datetime_to_isodate(current_datetime=datetime.now())

    def to_json(self):
        return dataclasses.asdict(self, dict_factory=factory)

    def __str__(self):
        return json.dumps(self.to_json())



from pandas import DataFrame

from database.Execution import Execution
from enums.MetadataColumns import MetadataColumns
from enums.Profile import Profile
from preprocessing.Preprocess import Preprocess
from utils.setup_logger import log


class PreprocessHsjd(Preprocess):
    def __init__(self, execution: Execution, data: DataFrame, metadata: DataFrame, profile: str):
        super().__init__(execution=execution, data=data, metadata=metadata, profile=profile)

        self.metadata = metadata
        self.mapping_full_name_to_var_name = {}

    def preprocess(self):
        log.info("pre-process HSJD data")
        log.info(self.data)
        log.info(self.data.columns)
        # 1. add patient IDs for the data file "Phenotypic_Table.xlsx"
        # (other files containing phenotypic data do already contain patient IDs, starting from 1)
        normalized_columns = [MetadataColumns.normalize_name(column_name=column) for column in self.data.columns]
        if self.profile == Profile.PHENOTYPIC and "Phenotypic_Table" in self.execution.current_filepath and self.execution.patient_id_column_name not in normalized_columns:
            self.data[self.execution.patient_id_column_name] = [i for i in range(1, len(self.data)+1)]
        log.info(self.data.columns)
        log.info(self.data)
